// System Includes
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <vector>

// XML-RPC includes
#include <xmlrpc-c/base.hpp>
#include <xmlrpc-c/client.hpp>

// Local Includes
#include "sci_app_client.hpp"

using std::cerr;
using std::cout;
using std::endl;
using std::string;
using std::vector;

namespace prc {

// Initialize the singleton instance to NULL.
ScienceApplicationClient* prc::ScienceApplicationClient::clientInstance = NULL;
const string prc::ScienceApplicationClient::GET_CHECKPOINT_RPC_NAME = "client.getCheckpoint";
const string prc::ScienceApplicationClient::SAVE_CHECKPOINT_RPC_NAME = "client.saveCheckpoint";

/**
 * @return Returns the singleton instance of the ScienceApplicationClient
 */
ScienceApplicationClient* ScienceApplicationClient::instance() {
	if (clientInstance == NULL)
		clientInstance = new ScienceApplicationClient();
	return clientInstance;
}

/**
 * Sets the checkpointing flag.  The getCheckpoint and saveCheckpoint
 * methods will only execute RPCs on the PRC client when this flag is 
 * set to true.  If the flag is set to false and the setCheckpoint 
 * method is called, it will do nothing.  If the flag is set to false 
 * and the getCheckpoint method is called, it will simply return 0.
 * @param doCheckpointing True to enable checkpointing, false to disable it.
 */
void ScienceApplicationClient::setCheckpointing(bool doCheckpointing) {
	this->doCheckpointing = doCheckpointing;
}

/**
 * Retrieves the last checkpoint from the project client.  The saveCheckpoint() function should 
 * have been called before this function.
 * @return Returns a vector<byte> containing the checkpoint data, or a vector with size = 0 if 
 * the server did not find any saved checkpoints.
 */
vector<byte> ScienceApplicationClient::getCheckpoint() {
	vector<byte> checkpoint;

	// Checkpointing is enabled.
	if (doCheckpointing == true) {
		// Build parameters and the XML-RPC.
		xmlrpc_c::paramList params;
		xmlrpc_c::rpcPtr xmlrpc(GET_CHECKPOINT_RPC_NAME, params);
	
		// Execute the RPC.
		try {
			xmlrpc_c::value result = executeRPC(xmlrpc);
	
			xmlrpc_c::value_bytestring dataStr = (xmlrpc_c::value_bytestring)result;
			checkpoint = dataStr.vectorUcharValue();
			if (checkpoint.size() > 0)
				cout << "ScienceApplicationClient::getCheckpoint: Found a checkpoint." << endl;
		} catch (...) {
			cerr << "Science Application: Error connecting to PRC Client." << endl;
			cerr << "Science Application: Shutting down..." << endl;
			exit(EXIT_FAILURE);
		}
	
		return checkpoint;
	}

	// Checkpointing is disabled.
	else {
		vector<byte> dummyVector(0);
		return dummyVector;
	}
}

/**
 * This function is a wrapper around saveCheckpoint(byte*,size_t).
 * It converts the given vector into a byte array, then saves the checkpoint.
 * @param data Vector containing the checkpoint data to save.
 * @return Returns true if the checkpoint was successfully written, false otherwise.
 */
bool ScienceApplicationClient::saveCheckpoint(const std::vector<byte>& data) {
	// Checkpointing is enabled.
	if (doCheckpointing == true) {
		cout << "ScienceApplication::writeCheckpoint: Checkpoint saved." << endl;
		xmlrpc_c::value_bytestring const dataParam(data);

		// Build parameters and the XML-RPC.
		xmlrpc_c::paramList params(1);
		params.add(dataParam);
		xmlrpc_c::rpcPtr xmlrpc(SAVE_CHECKPOINT_RPC_NAME, params);
	
		// Execute the RPC.
		bool success = false;
		try {
			xmlrpc_c::value result = executeRPC(xmlrpc);
			success = (bool) xmlrpc_c::value_boolean(result);
		} catch (...) {
			cerr << "Science Application: Error connecting to PRC Client." << endl;
			cerr << "Science Application: Shutting down..." << endl;
			exit(EXIT_FAILURE);
		}
	
		return success;
	}

	// Checkpointing is disabled.
	else
		return false;
}

} // end namespace prc
