// System Includes
#include <iostream>
#include <math.h>
#include <pthread.h>
#include <sched.h>
#include <vector>

// Local Includes
#include "common/common.hpp"
#include "sci_app_client.hpp"
#include "sci_data_proc.hpp"

using std::cout;
using std::endl;
using std::string;
using std::vector;

namespace prc {
/**
 * This method will be called by the RPC handlers.
 * The run method should execute the project-specific 
 * science algorithm on the given work unit and produce 
 * a result in the form of a vector<byte>.  The byte 
 * vector should be passed back in the return value 
 * of this method.
 * @param workUnitData The work unit data to process.
 * @return Returns the calculated result data.
 */
vector<byte> ScienceDataProcessor::run(vector<byte>& workUnitData) {
	// Note: It is necessary to split the vector into 2 vectors that each have 
	// sizeof(int) elements.  We do this using the iterator constructor.
	vector<byte>::iterator intPartition = workUnitData.begin() + sizeof(int);
	vector<byte> startVector(workUnitData.begin(), intPartition);
	vector<byte> endVector(intPartition, workUnitData.end());

	// Convert the byte vectors representing the two ints into their int values.
	int startRange = byteVectorToInt(startVector);
	int endRange = byteVectorToInt(endVector);

	cout << "Science Application: Start range is " << startRange << endl;
	cout << "Science Application: End range is " << endRange << endl;

	// This vector will store all computed primes, including both 
	// primes retrieved from a saved checkpoint as well as primes
	// computed during this run of the algorithm.
	vector<int> primes;

	// Check the client for any saved checkpoints.
	vector<byte> checkpoint = ScienceApplicationClient::instance()->getCheckpoint();
	
	// The client had a checkpoint.
	if (checkpoint.size() > 0) {
		// Extract the checkpoint data consisting of the adjusted start range
		// and all primes found the last time the science application ran.
		int newStartRange = byteVectorToInt(checkpoint);
			
		// Make sure the start range value is valid.
		if ((newStartRange >= 0) && (newStartRange < endRange)) {
			startRange = newStartRange;
			cout << "Science Application: New start range is " << newStartRange << endl;

			// Extract any primes that were found up to the last checkpoint.
			// NOTE: We skip the first integer because it is the start range, not a prime.
			for (vector<byte>::iterator begin = checkpoint.begin() + sizeof(int); begin < checkpoint.end(); begin += sizeof(int)) {
				vector<byte>::iterator end = begin + sizeof(int);
				vector<byte> primeBytes(begin, end);
				int prime = byteVectorToInt(primeBytes);
				primes.push_back(prime);
			}
		}
	}

	// Now find all primes

	// First, initialize a low-priority compute thread.
	pthread_attr_t threadAttrs;
	pthread_attr_init(&threadAttrs);
	struct sched_param threadSchedParam;
	threadSchedParam.sched_priority = sched_get_priority_min(SCHED_FIFO);
	pthread_attr_setschedparam(&threadAttrs, &threadSchedParam);

	// Start findPrimes() in the compute thread.
	int threadParams[2] = { startRange, endRange };
	pthread_create(&computeThread, &threadAttrs, findPrimes, (void*)threadParams);
	
	// Then, wait for the thread to complete.
	void* computeThreadResult;
	pthread_join(computeThread, &computeThreadResult);

	// When the thread has finished, cast the return value to a vector<int>*.
	vector<int>* foundPrimes = static_cast<vector<int>*>(computeThreadResult);
	
	// Copy all the primes found by findPrimes() into the primes vector.
	primes.insert(primes.end(), foundPrimes->begin(), foundPrimes->end());

	// Clean up
	delete foundPrimes;

	// Convert the primes vector into a byte vector.
	vector<byte> resultData;
	for (unsigned int ii = 0; ii < primes.size(); ii++) {
		vector<byte> primeBytes = intToByteVector(primes[ii]);
		resultData.insert(resultData.end(), primeBytes.begin(), primeBytes.end());
	}

	return resultData;
}

/**
 * Finds all prime numbers in the given interval.
 * @param range The range over which to search for prime numbers.
 * @return Returns a vector of all primes that were found.
 */
void* ScienceDataProcessor::findPrimes(void* range) {
	int* workRange = static_cast<int*>(range);
	int start = workRange[0];
	int end = workRange[1];
	vector<int>* primes = new vector<int>;

	cout << "Science Application: Testing Range [" << start << "," << end << "]" << endl;

	// This algorithm is not the most efficient, but it works.
	for (int ii = start; ii <= end; ii++) {
		bool isPrime = true;
		int maxSearch = (int)sqrt((double)ii);		// Only need to search up to 
								//the square root of the number.
		for (int jj = 2; jj <= maxSearch; jj++) {
			if (ii % jj == 0) {
				isPrime = false;
				break;
			}
		}
			
		if (isPrime == true) {
			cout << "Science Application: Found Prime: " << ii << endl;
			primes->push_back(ii);
		}
			
		// Write checkpoint after every CHECKPOINT_INTERVAL numbers checked.
		if ((ii % CHECKPOINT_INTERVAL == 0) && (ii + 1 < end)) {
			vector<byte> checkpointData;

			vector<byte> dataArr = intToByteVector(ii + 1);
			checkpointData.insert(checkpointData.end(), dataArr.begin(), dataArr.end());

			for (unsigned int nn = 0; nn < primes->size(); nn++) {
				dataArr = intToByteVector((*primes)[nn]);
				checkpointData.insert(checkpointData.end(), dataArr.begin(), dataArr.end());
			}

			ScienceApplicationClient::instance()->saveCheckpoint(checkpointData);
		}
	}
	
	return primes;
}

} // end namespace prc
