// System includes
#include <iostream>
#include <cstring>
#include <vector>

// Local includes
#include "common/common.hpp"
#include "result_validator.hpp"

using std::cout;
using std::endl;
using std::string;
using std::vector;

namespace prc {

/**
 * Validates a single result.
 * @param r The result to validate.
 * @return Returns true if the result is valid, false otherwise.
 */
bool ResultValidator::validateSingleResult(Result& r) {
	cout << "Result Validator: Validating Result " << r.resultID << endl;

	// We want to verify that each integer calculated by the science application is in the 
	// range given by the work unit, so first get the work unit from which this result was computed.
	vector<byte> workUnitData = ValidatorClient::instance()->getAssociatedWorkUnit(r.resultID);

	// Convert the byte vectors representing the two ints into their int values.
	vector<byte>::iterator intPartition = workUnitData.begin() + sizeof(int);
	vector<byte> startVector(workUnitData.begin(), intPartition);
	vector<byte> endVector(intPartition, workUnitData.end());
	int startRange = byteVectorToInt(startVector);
	int endRange = byteVectorToInt(endVector);

	cout << "Result Validator: Start range was " << startRange << endl;
	cout << "Result Validator: End range was " << endRange << endl;

	// Examine each computed integer and verify that it is in the range.
	bool allIntegersValid = true;
	int numChecked = 0;
	vector<byte>::iterator intStart = r.data.begin();
	vector<byte>::iterator intEnd = r.data.begin() + sizeof(int);
	// Keep looping while there are more integers to read.
	while (intEnd <= r.data.end()) {
		vector<byte> primeData(intStart, intEnd);
		int prime = byteVectorToInt(primeData);
		// Check whether the prime is in the range.
		if ((prime < startRange) || (prime > endRange)) {
			cout << "Result Validator: Detected invalid result data." << endl;
			cout << "Result Validator: " << prime << " is not in the work unit's range." << endl;
			allIntegersValid = false;
			break;
		}
		intStart += sizeof(int);
		intEnd += sizeof(int);
		numChecked++;
	}

	if (allIntegersValid == true)
		cout << "Result Validator: Found " << numChecked << " valid primes." << endl;
	
	return allIntegersValid;
}

/**
 * Selects a canonical result from a set of results.
 * @param results The set of valid results for a work unit.
 * @return Returns the result ID of the canonical result that was selected.
 */
string ResultValidator::selectCanonicalResult(vector<Result>& results) {
	// In this validator we always select the first result.
	return results[0].resultID;
}

/**
 * Validates a spot-check.
 * @param check The spot-check to validate.
 * @return Returns true if the spot-check is valid, false otherwise.
 */
bool ResultValidator::validateSpotCheck(SpotCheck& check) {
	// For this project we don't have floating point values or any nondeterminism, 
	// so all results should be exactly the same.  This method does a byte-by-byte 
	// comparison of the spot-check to the accepted result.
	bool validSpotCheck = true;
		
	// If the length of the two byte vectors is not the same, the spot-check is invalid.
	if (check.clientData.size() != check.acceptedData.size())
		validSpotCheck = false;
		
	// Do a byte-by-byte comparison of the two byte arrays.
	else {
		for (unsigned int ii = 0; ii < check.clientData.size(); ii++) {
			if (check.clientData[ii] != check.acceptedData[ii]) {
				validSpotCheck = false;
				break;
			}	
		}
	}

	return validSpotCheck;
}

} // end namespace prc
