// System includes
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <vector>

// Local includes
#include "common/common.hpp"
#include "generator.hpp"
#include "work_gen_client.hpp"

using std::cout;
using std::endl;
using std::string;
using std::stringstream;
using std::vector;

/**
 * The WorkUnitGenerator contains the project-specific 
 * code to generate work units.  It uses the WorkGeneratorClient
 * to send them to the project server.
 * @author James Baldassari
 */
namespace prc {

/**
 * The default constructor calls recoverState()
 * to initialize the work unit ID and the starting
 * and ending ranges for the next work unit to be 
 * generated.
 */
WorkUnitGenerator::WorkUnitGenerator() {
	recoverState();
}

/**
 * Retrieves the last work unit from the server
 * and uses it to initialize the WorkUnitGenerator
 * so that it can resume generating where it left off.
 */
void WorkUnitGenerator::recoverState() {
	// Get the last work unit
	try {
		WorkUnit wu = WorkGeneratorClient::instance()->getLastWorkUnit();
		// We received a work unit from the server.
		// Inspect the last work unit to determine how to generate
		// the next one.

		// Our work units are named WU_1, WU_2, WU_3, ...
		// So the number begins at index 3 of the task ID.
		// Determine the next task ID by adding 1 to 
		// the number that was used in the last task ID.
		nextWorkUnitID = atoi(wu.workUnitID.c_str() + 3) + 1;

		// Determine the next starting range by adding 1
		// to the last ending range.
		int end = byteVectorToInt(wu.data);
		cout << "Last end range was " << end << endl;
		nextStartRange = end + 1;
	} catch (string error) {
		// The server had zero work units.
		// This means that we should start generating
		// our first work unit.

		cout << "recoverState: Initializing the work unit generator for the first work unit." << endl;

		nextWorkUnitID = 1;
		nextStartRange = 2;	// No point in testing numbers < 2 ;)
	}

	// Set the end range relative to the start range.
	nextEndRange = nextStartRange + RANGE_LENGTH;
}

/**
 * Generates the next work unit.
 * @return Returns a pointer to a vector<byte> that contains the work unit data.
 */
vector<byte>* WorkUnitGenerator::generateWorkUnit() {
	cout << "WorkUnitGenerator.generateWorkUnit: nextStartRange = " << nextStartRange << endl;
	cout << "WorkUnitGenerator.generateWorkUnit: nextEndRange = " << nextEndRange << endl;
	vector<byte>* wu = new vector<byte>;

	// Convert start and end ranges to vector<byte> objects.
	vector<byte> startRange = intToByteVector(nextStartRange);
	vector<byte> endRange = intToByteVector(nextEndRange);
	
	// Insert the two vectors into the wu vector.
	wu->insert(wu->begin(), startRange.begin(), startRange.end());
	wu->insert(wu->end(), endRange.begin(), endRange.end());

	// Update the ranges.
	nextStartRange = nextEndRange + 1;
	nextEndRange = nextStartRange + RANGE_LENGTH;

	return wu;
}

/**
 * Generates and sends a given number of work units to the project server.
 * @param numWorkUnits The number of work units to generate and send.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkUnitGenerator::generateAndSend(int numWorkUnits) {
	cout << "Generating " << numWorkUnits << " new work units, then sending them to the server." << endl;
	size_t numWaiting = 0;

	for (int ii = 0; ii < numWorkUnits; ii++) {
		// Build the next work unit ID.
		stringstream idStream(std::ios::out);
		idStream << "WU_" << nextWorkUnitID;
		string workUnitID = idStream.str();
		nextWorkUnitID++;

		// Generate the next work unit data.
		vector<byte>* wu = generateWorkUnit();

		// Send the work unit to the server.
		numWaiting = WorkGeneratorClient::instance()->sendWorkUnit(
			workUnitID,
			*wu,
			DEFAULT_PRIORITY,
			DEFAULT_POINTS,
			DEFAULT_EXPIRATION_TIME);

		// Clean up
		delete wu;
	}

	return numWaiting;
}

} // end namespace prc
