// System Includes
#include <cstdlib>
#include <ctime>
#include <iostream>

// XML-RPC includes
#include <xmlrpc-c/base.hpp>
#include <xmlrpc-c/client.hpp>

// Local Includes
#include "common/common.hpp"
#include "work_gen_client.hpp"

using std::cerr;
using std::cout;
using std::endl;
using std::string;
using std::time_t;
using std::vector;

namespace prc {

// Initialize the singleton instance to NULL.
WorkGeneratorClient* prc::WorkGeneratorClient::clientInstance = NULL;
const string prc::WorkGeneratorClient::GET_LAST_WU_RPC_NAME = "server.getLastWorkUnit";
const string prc::WorkGeneratorClient::GET_NUM_WU_RPC_NAME = "server.getNumWorkUnits";
const string prc::WorkGeneratorClient::SEND_WU_RPC_NAME = "server.addWorkUnit";

/**
 * @return Returns the singleton instance of the WorkGeneratorClient
 */
WorkGeneratorClient* WorkGeneratorClient::instance() {
	if (clientInstance == NULL)
		clientInstance = new WorkGeneratorClient();
	return clientInstance;
}

/**
 * Executes the server.addWorkUnit RPC with the given parameter list.
 * @param params The parameters to the server.addWorkUnit RPC.
 * @param errorMessage The message to be displayed if an error occurs.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(xmlrpc_c::paramList& params, const string& errorMessage) {
	size_t numWaiting = 0;

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(SEND_WU_RPC_NAME, params);

	// Execute the RPC.
	try {
		xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);
		numWaiting = (size_t) ((xmlrpc_c::value_int)result);
	} catch (string error) {
		// Some unexpected error occurred, such as the given RPC was not found.
		cerr << "ERROR: The server returned: " << error << endl;
		cerr << "Work generator exiting..." << endl;
		exit(EXIT_FAILURE);
	}

	// Convert the result into a size_t.
	return numWaiting;
}

/**
 * Retrieves the last work unit that was generated from the server.
 * @return Returns a WorkUnit struct that contains all work unit 
 * information returned by the server.
 */
WorkUnit WorkGeneratorClient::getLastWorkUnit() {
	xmlrpc_c::paramList params;
	xmlrpc_c::rpcPtr xmlrpc(GET_LAST_WU_RPC_NAME, params);
	string errorMessage = "Error while retrieving the last work unit";
	WorkUnit wu;

	// Execute the RPC.
	try {
		xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);
	
		// Convert the result to a vector of xmlrpc_c::value objects.
		vector<xmlrpc_c::value> rpcValues = xmlrpc_c::value_array(result).vectorValueValue();
	
		// Add all of the elements of the vector to the WorkUnit,
		// stripping the xmlrpc-c wrapper types.
	
		// The first element is a string representing the work unit ID.
		wu.workUnitID = (string)((xmlrpc_c::value_string)rpcValues[0]);
	
		// The second element is the work unit data.
		// This will be stored in a vector.
		wu.data = (vector<byte>)((xmlrpc_c::value_bytestring)rpcValues[1]).vectorUcharValue();
	
		// The third element is the creation date of the work unit.
		wu.creationDate = (time_t)((xmlrpc_c::value_datetime)rpcValues[2]);
	
		// The fourth element is the point value of the work unit.
		wu.points = (int)((xmlrpc_c::value_int)rpcValues[3]);
	
		// The fifth element is the priority of the work unit.
		wu.priority = (int)((xmlrpc_c::value_int)rpcValues[4]);
	} catch (string error) {
		// The server did not have any work units,
		// or an error occurred such as the given RPC was not found.
		//cout << "The server returned: " << error << endl;
		cout << "getLastWorkUnit: The server did not have any work units." << endl;
		// Rethrow the exception
		throw error;
	}

	return wu;
}

/**
 * Sends a new work unit to the server.
 * @param taskID The ID of the work unit.
 * @param data Vector<byte> containing the work unit data.
 * @param priority The priority of the work unit.  The minimum 
 * priority is zero, and higher values correspond to higher 
 * priorities.
 * @param points The number of points to give a volunteer upon 
 * returning a valid result for this work unit.
 * @param expirationTime The number of seconds until this work unit
 * should expire.  If the work unit is not retired before this amount
 * of time elapses, the work unit will be recycled into the ingress 
 * queue and will be distributed to other volunteers.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(const string taskID, const vector<byte> data,
					int priority, int points, double expirationTime) {
	string errorMessage = "Error while sending a work unit";
	xmlrpc_c::paramList params;

	// Build the parameters.
	params.add(xmlrpc_c::value_string(taskID));
	params.add(xmlrpc_c::value_bytestring(data));
	params.add(xmlrpc_c::value_int(priority));
	params.add(xmlrpc_c::value_int(points));
	params.add(xmlrpc_c::value_double(expirationTime));

	return sendWorkUnit(params, errorMessage);
}

/**
 * Sends a new work unit to the server.
 * @param taskID The ID of the work unit.
 * @param data Vector<byte> containing the work unit data.
 * @param priority The priority of the work unit.  The minimum 
 * priority is zero, and higher values correspond to higher 
 * priorities.
 * @param points The number of points to give a volunteer upon 
 * returning a valid result for this work unit.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(const string taskID, const vector<byte> data,
					int priority, int points) {
	string errorMessage = "Error while sending a work unit";
	xmlrpc_c::paramList params;

	// Build the parameters.
	params.add(xmlrpc_c::value_string(taskID));
	params.add(xmlrpc_c::value_bytestring(data));
	params.add(xmlrpc_c::value_int(priority));
	params.add(xmlrpc_c::value_int(points));

	return sendWorkUnit(params, errorMessage);
}

/**
 * Sends a new work unit to the server.
 * @param data Vector<byte> containing the work unit data.
 * @param priority The priority of the work unit.  The minimum 
 * priority is zero, and higher values correspond to higher 
 * priorities.
 * @param points The number of points to give a volunteer upon 
 * returning a valid result for this work unit.
 * @param expirationTime The number of seconds until this work unit
 * should expire.  If the work unit is not retired before this amount
 * of time elapses, the work unit will be recycled into the ingress 
 * queue and will be distributed to other volunteers.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(const vector<byte> data, int priority, int points, 
					double expirationTime) {
	string errorMessage = "Error while sending a work unit";
	xmlrpc_c::paramList params;

	// Build the parameters.
	params.add(xmlrpc_c::value_bytestring(data));
	params.add(xmlrpc_c::value_int(priority));
	params.add(xmlrpc_c::value_int(points));
	params.add(xmlrpc_c::value_double(expirationTime));

	return sendWorkUnit(params, errorMessage);
}

/**
 * Sends a new work unit to the server.
 * @param data Vector<byte> containing the work unit data.
 * @param priority The priority of the work unit.  The minimum 
 * priority is zero, and higher values correspond to higher 
 * priorities.
 * @param points The number of points to give a volunteer upon 
 * returning a valid result for this work unit.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(const vector<byte> data, int priority, int points) {
	string errorMessage = "Error while sending a work unit";
	xmlrpc_c::paramList params;

	// Build the parameters.
	params.add(xmlrpc_c::value_bytestring(data));
	params.add(xmlrpc_c::value_int(priority));
	params.add(xmlrpc_c::value_int(points));

	return sendWorkUnit(params, errorMessage);
}

/**
 * Sends a new work unit to the server.
 * @param data Vector<byte> containing the work unit data.
 * @return Returns the number of clients that are blocked waiting for work units.
 */
size_t WorkGeneratorClient::sendWorkUnit(const vector<byte> data) {
	string errorMessage = "Error while sending a work unit";
	xmlrpc_c::paramList params;

	// Build the parameters.
	params.add(xmlrpc_c::value_bytestring(data));

	return sendWorkUnit(params, errorMessage);
}

/**
 * Retrieves the number of ingress (new) work units in the transitioner.
 * As this number approaches zero, more work units should be generated.
 * @return The number of ingress work units in the transitioner.
 */
size_t WorkGeneratorClient::getNumWorkUnits() {
	string errorMessage = "Error while requesting the number of ingress work units";
	xmlrpc_c::paramList params;

	// Build the RPC.
	xmlrpc_c::rpcPtr xmlrpc(GET_NUM_WU_RPC_NAME, params);

	// Execute the RPC.
	xmlrpc_c::value result = executeRPCUntilSuccess(xmlrpc, errorMessage);

	// Convert the result into a size_t.
	return (size_t) ((xmlrpc_c::value_int)result);
}

} // end namespace prc
