package edu.wpi.prc.example.java.generator;

import java.util.Date;
import java.util.Vector;

import edu.wpi.prc.example.java.common.Client;
import edu.wpi.prc.example.java.common.RPCHandlerException;

/**
 * The WorkGeneratorClient is an XML-RPC client with 
 * added functionality that is needed by the work unit generator.
 * It has methods for sending new work units to the server, retrieving 
 * the last work unit that was added, and retrieving the number of new
 * (ingress) work units that the server has.
 * @author James Baldassari
 */
public class WorkGeneratorClient extends Client {
	private static final int EXIT_ERROR = 1;
	private static final String GET_LAST_WU_RPC_NAME = "server.getLastWorkUnit";
	private static final String GET_NUM_WU_RPC_NAME = "server.getNumWorkUnits";
	private static final String SEND_WU_RPC_NAME = "server.addWorkUnit";
	private static final String DEFAULT_SERVER_ADDRESS = "127.0.0.1:2080";
	private static WorkGeneratorClient client = null;
	
	/**
	 * Constructor specifying the address of the project server.
	 * This constructor is private because WorkGeneratorClient is a singleton.
	 * @param serverAddress The address (host:port) of the project server.
	 */
	private WorkGeneratorClient(String serverAddress) {
		super(serverAddress);
	}
	
	/**
	 * Executes the server.addWorkUnit RPC with the given parameter list.
	 * @param params The parameters to the server.addWorkUnit RPC.
	 * @param errorMessage The message to be displayed if an error occurs.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	private int sendWorkUnit(Vector params, String errorMessage) {
		int numWaiting = -1;
		
		try {
			Object rpcValue = executeRPCUntilSuccess(SEND_WU_RPC_NAME, params, errorMessage);
			numWaiting = (Integer)rpcValue;
		} catch (ClassCastException e) {
			System.out.println("The server returned an unexpected error: " + e);
			shutdown();
		} catch (RPCHandlerException e) {
			System.out.println("The server returned an unexpected error: " + e);
			shutdown();
		}
		
		return numWaiting;
	}
	
	/**
	 * Retrieves the last work unit that was generated from the server.
	 * @return Returns a WorkUnit object containing all information 
	 * returned by the server.
	 */
	public WorkUnit getLastWorkUnit() throws RPCHandlerException {
		String errorMessage = "Error while retrieving the last work unit";
		WorkUnit wu = null;
		try {
			Object rpcValue = executeRPCUntilSuccess(GET_LAST_WU_RPC_NAME, new Vector(), errorMessage);
			Vector wuVector = (Vector)rpcValue;
			
			// Extract the individual elements from the vector.
			String workUnitID = (String)wuVector.get(0);
			byte[] workUnitData = (byte[])wuVector.get(1);
			Date creationDate = (Date)wuVector.get(2);
			int points = (Integer)wuVector.get(3);
			int priority = (Integer)wuVector.get(4);
			
			// Store these values in a WorkUnit for convenient access.
			wu = new WorkUnit(workUnitID, workUnitData, creationDate, points, priority);
		} catch (ClassCastException e) {
			//System.out.println("The server returned: " + e);
			System.out.println("getLastWorkUnit: The server did not have any work units.");
			// Rethrow the Exception, which will have to be caught by the WorkUnitGenerator.
			throw new RPCHandlerException(e.getMessage());
		} catch (RPCHandlerException e) {
			//System.out.println("The server returned: " + e);
			System.out.println("getLastWorkUnit: The server did not have any work units.");
			// Rethrow the Exception, which will have to be caught by the WorkUnitGenerator.
			throw e;
		}
		
		return wu;
	}
	
	/**
	 * Sends a new work unit to the server.
	 * @param workUnitID The ID of the work unit.
	 * @param data A byte array containing the work unit data.
	 * @param priority The priority of the work unit.  The minimum 
	 * priority is zero, and higher values correspond to higher 
	 * priorities.
	 * @param points The number of points to give a volunteer upon 
	 * returning a valid result for this work unit.
	 * @param expirationTime The number of seconds until this work unit
	 * should expire.  If the work unit is not retired before this amount
	 * of time elapses, the work unit will be recycled into the ingress 
	 * queue and will be distributed to other volunteers.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	@SuppressWarnings("unchecked")
	public int sendWorkUnit(String workUnitID, byte[] data, int priority, int points, double expirationTime) {
		String errorMessage = "Error while sending a work unit";
		
		// Build parameter list.
		Vector params = new Vector(5);
		params.add(workUnitID);
		params.add(data);
		params.add(priority);
		params.add(points);
		params.add(expirationTime);
		
		return sendWorkUnit(params, errorMessage);
	}
	
	/**
	 * Sends a new work unit to the server.
	 * @param workUnitID The ID of the work unit.
	 * @param data A byte array containing the work unit data.
	 * @param priority The priority of the work unit.  The minimum 
	 * priority is zero, and higher values correspond to higher 
	 * priorities.
	 * @param points The number of points to give a volunteer upon 
	 * returning a valid result for this work unit.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	@SuppressWarnings("unchecked")
	public int sendWorkUnit(String workUnitID, byte[] data, int priority, int points) {
		String errorMessage = "Error while sending a work unit";
		
		// Build parameter list.
		Vector params = new Vector(4);
		params.add(workUnitID);
		params.add(data);
		params.add(priority);
		params.add(points);
		
		return sendWorkUnit(params, errorMessage);
	}
	
	/**
	 * Sends a new work unit to the server.
	 * @param data A byte array containing the work unit data.
	 * @param priority The priority of the work unit.  The minimum 
	 * priority is zero, and higher values correspond to higher 
	 * priorities.
	 * @param points The number of points to give a volunteer upon 
	 * returning a valid result for this work unit.
	 * @param expirationTime The number of seconds until this work unit
	 * should expire.  If the work unit is not retired before this amount
	 * of time elapses, the work unit will be recycled into the ingress 
	 * queue and will be distributed to other volunteers.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	@SuppressWarnings("unchecked")
	public int sendWorkUnit(byte[] data, int priority, int points, double expirationTime) {
		String errorMessage = "Error while sending a work unit";
		
		// Build parameter list.
		Vector params = new Vector(4);
		params.add(data);
		params.add(priority);
		params.add(points);
		params.add(expirationTime);
		
		return sendWorkUnit(params, errorMessage);
	}
	
	/**
	 * Sends a new work unit to the server.
	 * @param data A byte array containing the work unit data.
	 * @param priority The priority of the work unit.  The minimum 
	 * priority is zero, and higher values correspond to higher 
	 * priorities.
	 * @param points The number of points to give a volunteer upon 
	 * returning a valid result for this work unit.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	@SuppressWarnings("unchecked")
	public int sendWorkUnit(byte[] data, int priority, int points) {
		String errorMessage = "Error while sending a work unit";
		
		// Build parameter list.
		Vector params = new Vector(3);
		params.add(data);
		params.add(priority);
		params.add(points);
		
		return sendWorkUnit(params, errorMessage);
	}
	
	/**
	 * Sends a new work unit to the server.
	 * @param data A byte array containing the work unit data.
	 * @return Returns the number of clients that are blocked waiting for work units.
	 */
	@SuppressWarnings("unchecked")
	public int sendWorkUnit(byte[] data) {
		String errorMessage = "Error while sending a work unit";
		
		// Build parameter list.
		Vector params = new Vector(1);
		params.add(data);
		
		return sendWorkUnit(params, errorMessage);
	}
	
	/**
	 * Retrieves the number of ingress (new) work units in the transitioner.
	 * As this number approaches zero, more work units should be generated.
	 * @return The number of ingress work units in the transitioner.
	 */
	public int getNumWorkUnits() {
		int numWorkUnits = -1;
		String errorMessage = "Error while requesting the number of ingress work units";
		
		try {
			Object rpcResult = executeRPCUntilSuccess(GET_NUM_WU_RPC_NAME, new Vector(), errorMessage);
			numWorkUnits = (Integer)rpcResult;
		} catch (ClassCastException e) {
			System.out.println("The server returned an unexpected error: " + e);
			shutdown();
		} catch (RPCHandlerException e) {
			System.out.println("The server returned an unexpected error: " + e);
			shutdown();
		}
		
		return numWorkUnits;
	}
	
	/**
	 * Exits the entire process because an unrecoverable error occurred.
	 */
	private void shutdown() {
		System.err.println("Science Application: Error connecting to PRC Client.");
		System.err.println("Science Application: Shutting down...");
		System.exit(EXIT_ERROR);
	}
	
	/**
	 * @return Returns the singleton instance of this class.
	 */
	public static WorkGeneratorClient instance() {
		if (client == null)
			client = new WorkGeneratorClient(DEFAULT_SERVER_ADDRESS);
		return client;
	}
}
