package edu.wpi.prc.example.java.generator;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import edu.wpi.prc.example.java.common.RPCHandlerException;

/**
 * The WorkUnitGenerator contains the project-specific 
 * code to generate work units.  It uses the WorkGeneratorClient
 * to send them to the project server.
 * @author James Baldassari
 */
public class WorkUnitGenerator {
	private static final int RANGE_LENGTH = 1000000;					// How many numbers should be checked in each work unit?
	private static final int PRIORITY = 0;							// The PRIORITY of the work units to generate (0 is the default).
	private static final int POINTS = 1;								// The number of POINTS these work units are worth (1 is the default).
	private static final double EXPIRATION_TIME = 900;				// The number of seconds before each work unit expires (1 day is the default).
	private static final int INT_SIZE = Integer.SIZE / 8;			// Size in bytes of an integer.
	private int workUnitID = 0;										// The ID of the next work unit to be generated.  Used in the addWorkUnit RPC.
	private int nextStartRange;										// Where should the next range start?					
	private int nextEndRange;										// Where should the next range end?
	
	/**
	 * Default constructor.
	 * Gets last work unit generated via XML-RPC so 
	 * that it can resume generating from that point.
	 */
	public WorkUnitGenerator() {
		nextStartRange = recoverStartRange();
		System.out.println("WorkUnitGenerator: Setting nextStartRange to " + nextStartRange);
		nextEndRange = nextStartRange + RANGE_LENGTH;
	}
	
	/**
	 * If the work unit generator were shut down, then started back up, it would need to determine the last 
	 * work unit that was generated so that it would know where to resume generating new ranges.  This 
	 * method asks the server for the last work unit that was generated.  If the server throws an Exception, 
	 * it means that no work units had ever been generated, in which case the work unit generator is starting 
	 * for the first time.
	 */
	private int recoverStartRange() {
		WorkUnit wu = null;
		int nextStart = 2;			// The next start range.

		try {
			// Get the last work unit from the server.
			wu = WorkGeneratorClient.instance().getLastWorkUnit();

			// NOTE: If no Exception was thrown by this point, the server returned a work unit.
			
			// Initialize the next work unit ID by getting the last work unit ID and increasing it by one.
			workUnitID = Integer.parseInt(wu.getWorkUnitID().substring(wu.getWorkUnitID().lastIndexOf("_") + 1)) + 1;
			
			// We are interested in getting the last end range that was issued so that we can resume from there.
			byte[] data = wu.getData();
			
			// The 2nd integer in the byte array is the end range.
			byte[] endRange = new byte[INT_SIZE];
			for (int ii = 0; ii < INT_SIZE; ii++)	// Copy the last INT_SIZE bytes in data to the endRange array.
				endRange[ii] = data[ii+INT_SIZE];
			
			// Convert the byte array to an integer.
			ByteArrayInputStream bais = new ByteArrayInputStream(endRange);
			DataInputStream dis = new DataInputStream(bais);
			try {
				int end = dis.readInt();
				System.out.println("WorkUnitGenerator.recoverStartRange: Last end range was: " + end);
				nextStart = end + 1;	// The next start range will be the last end range plus one.
			} catch (IOException e) {
				// Error occured.  0 will be returned.
				e.printStackTrace();
				System.err.println("WorkUnitGenerator.recoverStartRange: Error converting byte array to integer.  Starting from 0.");
			}
		} catch (RPCHandlerException e) {
			// The server had zero work units.
			// This means that we should start generating
			// our first work unit.
			// workUnitID will be incremented when the work unit is 
			// generated, so the first work unit will have ID 1, not 0.
			workUnitID = 0;
		}

		return nextStart;
	}
	
	/**
	 * Generates the next work unit. 
	 * @return Returns a byte array representing the generated work unit.
	 */
	@SuppressWarnings("unchecked")
	private byte[] generateWorkUnit() {
		System.out.println("WorkUnitGenerator.generateWorkUnit: nextStartRange = " + nextStartRange);
		System.out.println("WorkUnitGenerator.generateWorkUnit: nextEndRange = " + nextEndRange);
		
		// The work unit data.
		byte[] data = null;
		
		// Convert the start and end ranges to byte arrays.
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		DataOutputStream dos = new DataOutputStream(baos);
		try {
			// Convert the start range integer.
			dos.writeInt(nextStartRange);
			dos.writeInt(nextEndRange);
			dos.flush();
			data = baos.toByteArray();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// Increment the work unit ID.
		workUnitID++;
		
		// Reset the start and end ranges.
		nextStartRange = nextEndRange + 1;
		nextEndRange = nextStartRange + RANGE_LENGTH;
		
		return data;
	}
	
	/**
	 * Adds a new work unit to the server by calling the server.addWorkUnit XML-RPC.
	 * @param workUnit The work unit to send to the server.
	 */
	private void sendWorkUnit(byte[] workUnit) {
		// Build the work unit ID
		String id = "WU_" + workUnitID;
		// NOTE: There are several variants of the "WorkGeneratorClient.sendWorkUnit" method.
		// The simplest case only takes the data byte array.  This example 
		// specifies several parameters including ID, data, priority, point value, 
		// and expiration time.
		// See WorkGeneratorClient for other versions of the RPC.
		WorkGeneratorClient.instance().sendWorkUnit(id, workUnit, PRIORITY, POINTS, EXPIRATION_TIME);
	}
	
	/**
	 * Generates the specified number of work units, and sends them to the project server.
	 * @param numWorkUnits The number of work units to generate and send.
	 */
	public void generateAndSend(int numWorkUnits) {
		System.out.println("Generating " + numWorkUnits + " new work units, then sending them to the server.");
		for (int ii = 0; ii < numWorkUnits; ii++) {
			byte[] workUnit = generateWorkUnit();
			sendWorkUnit(workUnit);
		}
	}
}
