package edu.wpi.prc.example.java.science;

import java.io.IOException;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.example.java.common.Client;

/**
 * The ScienceApplicationClient is an XML-RPC client with 
 * added functionality that is needed by the science application.
 * This class has methods for saving and retrieving checkpoints.
 * @author James Baldassari
 */
public class ScienceApplicationClient extends Client {
	private static final int EXIT_ERROR = 1;
	private static final String DEFAULT_SERVER_ADDRESS = "127.0.0.1:2081";	// Default location of the PRC client's XML-RPC server.
	private static final String GET_CHECKPOINT_RPC_NAME = "client.getCheckpoint";
	private static final String SAVE_CHECKPOINT_RPC_NAME = "client.saveCheckpoint";
	private boolean doCheckpointing = true;												// Should the science application use checkpointing?
	private static ScienceApplicationClient client = null;								// The singleton instance.
	
	/**
	 * Default constructor.  This constructor is private
	 * because the ScienceApplicationClient is a singleton.
	 */
	private ScienceApplicationClient() {
		super(DEFAULT_SERVER_ADDRESS);
	}
	
	/**
	 * Enables or disables the use of checkpointing.
	 * Checkpointing should not be used when computing a spot-check result.
	 * @param doCheckpointing True to enable checkpointing, false to disable it.
	 */
	public void setCheckpointing(boolean doCheckpointing) {
		this.doCheckpointing = doCheckpointing;
	}
	
	/**
	 * Sends a checkpoint to the project client, which then writes it out to disk.
	 * If checkpointing is disabled, no action is taken.
	 * @param data The checkpoint data to be sent. 
	 * @return Returns true if the checkpoint was successfully written; returns false
	 * if an error occurred or checkpointing was disabled.
	 */
	@SuppressWarnings("unchecked")
	public boolean saveCheckpoint(byte[] data) {
		if (doCheckpointing == true) {
			// Checkpointing is enabled.
			Vector params = new Vector<byte[]>();
			params.add(data);
			boolean success = false;
			try {
				// Execute the RPC.
				Object returnValue = executeRPC(SAVE_CHECKPOINT_RPC_NAME, params);
				// Get the return value.
				success = ((Boolean)returnValue).booleanValue();
				System.out.println("ScienceApplicationClient.saveCheckpoint: Checkpoint saved.");
			} catch (ClassCastException e) {
				System.out.println("The client returned an unexpected error: " + e);
				shutdown();
			} catch (XmlRpcException e) {
				System.err.println("ScienceApplicationClient.saveCheckpoint: Received an Exception from the client while writing a checkpoint: " + e);
				shutdown();
			} catch (IOException e) {
				System.err.println("ScienceApplicationClient.saveCheckpoint: Received an Exception from the client while writing a checkpoint: " + e);
				shutdown();
			}
			return success;
		}
		
		else {
			// Checkpointing is disabled.
			return false;
		}
	}
	
	/**
	 * Retrieves the last checkpoint from the project client.  The saveCheckpoint() method should 
	 * have been called before this method.  If checkpointing is disabled, no action is taken.
	 * @return Returns the checkpoint data returned by the client; returns null if no checkpoint was found or 
	 * if checkpointing was disabled.
	 */
	public byte[] getCheckpoint() {
		if (doCheckpointing == true) {
			// Checkpointing is enabled.
			System.out.println("ScienceApplicationClient.getCheckpoint: Retrieving checkpoint.");
			Vector params = new Vector<byte[]>();
			byte[] checkpoint = null;
			try {
				// Execute the RPC.
				Object returnValue = executeRPC(GET_CHECKPOINT_RPC_NAME, params);
				// Get the return value.
				checkpoint = (byte[]) returnValue;
				if (checkpoint.length > 0)
					System.out.println("ScienceApplicationClient.getCheckpoint: Found a checkpoint.");
				else
					System.out.println("ScienceApplicationClient.getCheckpoint: No checkpoint found.");
			} catch (ClassCastException e) {
				System.out.println("The client returned an unexpected error: " + e);
				shutdown();
			}  catch (XmlRpcException e) {
				System.err.println("ScienceApplicationClient.getCheckpoint: Received an Exception from the client while retrieving a checkpoint: " + e);
				shutdown();
			} catch (IOException e) {
				System.err.println("ScienceApplicationClient.getCheckpoint: Received an Exception from the client while retrieving a checkpoint: " + e);
				shutdown();
			}
			return checkpoint;
		}
		
		else {
			// Checkpointing is disabled.
			return new byte[0];
		}
	}
	
	/**
	 * Exits the entire process because an unrecoverable error occurred.
	 */
	private void shutdown() {
		System.err.println("Science Application: Error connecting to PRC Client.");
		System.err.println("Science Application: Shutting down...");
		System.exit(EXIT_ERROR);
	}
	
	/**
	 * @return Returns the singleton instance of the ScienceApplicationClient class.
	 */
	public static ScienceApplicationClient instance() {
		if (client == null)
			client = new ScienceApplicationClient();
		return client;
	}
}
