package edu.wpi.prc.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.util.Properties;
import java.util.Vector;

import org.apache.xmlrpc.WebServer;
import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.persistence.InvalidLogPathException;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.task.NonPersistentResult;
import edu.wpi.prc.task.NonPersistentWorkUnit;

/**
 * The PRC client acts as a proxy between the PRC server and the science application.
 * @author James Baldassari
 */
public class PRCClient implements Runnable {
	private static PRCClient prcClient = null;
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private static final String LOG_PATH = "log" + File.separator + "client.out.log";
	private static final String ERROR_LOG_PATH = "log" + File.separator + "client.err.log";
	private static final String CONFIG_FILE_PATH = "cfg" + File.separator + "client.properties";
	private static final String WORK_UNIT_ID_PATH = "data" + File.separator + "workunit.id";
	private static final String WORK_UNIT_DATA_PATH = "data" + File.separator + "workunit.dat";
	public static final String CHECKPOINT_PATH = "data" + File.separator + "checkpoint.dat"; 
	private static final int SERVER_RETRY_DELAY = 900000;	// Wait 15 minutes before retrying an RPC call to the server.
	private static final int SCIAPP_RETRY_DELAY = 30000;		// Wait 30 seconds before retrying an RPC call to the science app.
	private static final int STARTUP_DELAY = 10000;			// Wait 10 seconds for science application to start.
	private Thread clientThread = null;
	private Project project = null;
	private WebServer xmlrpcServer = null;
	private XmlRpcClient xmlrpcPRCServerClient = null;
	private XmlRpcClient xmlrpcScienceAppClient = null;
	private int clientPort;
	private String serverURI = null;
	private String scienceAppURI = null;
	private boolean isRunning = false;
	private String userName = null;
	private String eMailAddress = null;
	
	/**
	 * Private constructor prevents explcit instantiation of the PRCClient class.
	 * Initializes the client componenets.
	 * @throws MissingConfigurationFileException
	 */
	private PRCClient() throws MissingConfigurationFileException {
		clientThread = Thread.currentThread();
		configureClient();
		initXmlRpcServer();
		initXmlRpcClient();
		readClientConfigurationFile();
	}
	
	/**
	 * Reads configuration from files in the cfg directory.
	 * @throws MissingConfigurationFileException
	 */
	private void configureClient() throws MissingConfigurationFileException {
		project = Project.readFromFile();
		serverURI = "http://" + project.getServerURI() + "/RPC2";
		scienceAppURI = "http://localhost:" + project.getScienceAppPort() + "/RPC2";
		clientPort = project.getClientPort();
		try {
			// Redirect standard error to the error log file.
			System.setErr(new PrintStream(new FileOutputStream(ERROR_LOG_PATH)));
		} catch (FileNotFoundException e) {
			
		}
	}
	
	/**
	 * Reads the volunteer's user name and e-mail address from the configuration file.
	 * If the configuration file does not exist, the user is prompted for the required 
	 * information, and then the configuration file is saved to disk.
	 */
	private void readClientConfigurationFile() {
		try {
			InputStream is;
			is = new FileInputStream(CONFIG_FILE_PATH);
			Properties props = new Properties();
			props.load(is);
			
			// Read the properties from the file
			String user = props.getProperty("user_name", "");
			String eMail = props.getProperty("email_address", "");
			
			boolean isConfigured = false;
			if (!user.equals("")) {
				isConfigured = true;
				this.userName = user;
			}
			if (!eMail.equals("")) {
				isConfigured = true;
				this.eMailAddress = eMail;
			}
			
			if (isConfigured == false) {
				getConfigurationFromUser();
			}
		} catch (FileNotFoundException e) {
			getConfigurationFromUser();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Initializes and starts the XML-RPC server.
	 */
	private void initXmlRpcServer() {		
		// Make sure that the server URI has been configured before proceeding.
		if (serverURI == null) {
			System.out.println("Fatal error: The server URI is not configured.");
			System.err.println("Fatal error: The server URI is not configured.");
			System.exit(EXIT_ERROR);
		}
		
		else {
			try {
				// Instantiate the web server, configured to run on localhost.
				xmlrpcServer = new WebServer(clientPort, InetAddress.getByName("localhost"));
				
				// Register RPC handler.
				xmlrpcServer.addHandler("client", new ClientRPCHandler(this));
				
			} catch (Exception e) {
				SysLogger.println("Error initializing the XML-RPC server: " + e);
				System.err.println("Error initializing the XML-RPC server: " + e);
				System.exit(EXIT_ERROR);
			}
		}
	}
	
	/**
	 * Initializes the XML-RPC clients used for communicating with the project server 
	 * and science application.
	 */
	private void initXmlRpcClient() {
		try {
			xmlrpcPRCServerClient = new XmlRpcClient(serverURI);
			xmlrpcScienceAppClient = new XmlRpcClient(scienceAppURI);
		} catch (MalformedURLException e) {
			System.err.println("Invalid URL used to initialize the XML-RPC client: " + e);
			System.exit(EXIT_ERROR);
		}
	}
	
	/**
	 * Gets user name and e-mail address from user, then writes the configuration file.
	 */
	private void getConfigurationFromUser() {
		System.out.print("This installation needs to be configured before you can begin");
		System.out.print(" contributing to this public resource computing project.");
		System.out.print("  Please indicate whether you are a new or existing user by");
		System.out.println(" entering the number corresponding with your choice.");
		System.out.println("1. Existing user.");
		System.out.println("2. New user.");
		System.out.print("Please enter either 1 or 2: ");
		
		boolean validInput = false;
		int userType = 0;
		String user = null;
		String eMail = null;
		BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
		
		// Get user type.
		while(validInput == false) {
			String choice = null;
			try {
				choice = reader.readLine();
				userType = Integer.parseInt(choice);
				if ((userType == 1) || (userType == 2))
					validInput = true;
			} catch (IOException e) {
				//e.printStackTrace();
			} catch (NumberFormatException e) {
				System.out.println("Invalid input: \"" + choice + "\"");
				System.out.println("Please enter either 1 or 2.");
			}
		}
		
		// Get user name.
		boolean validUser = false;
		while (validUser == false) {
			validInput = false;
			while (validInput == false) {
				System.out.print("Enter user name: ");
				try {
					user = reader.readLine();
					//if (user.equals("") || !user.matches("\\W+"))
					if (user.equals("") || !user.matches("[a-zA-Z1-9]+"))
						System.out.println("User names must be at least one character long and can consist of only letters and numbers.");
					else
						validInput = true;
				} catch (IOException e) {

				}
			}
			
			// Get e-mail address.
			validInput = false;
			while (validInput == false) {
				try {
					System.out.print("Enter e-mail address: ");
					eMail = reader.readLine();
					if (eMail.equals("") || !eMail.matches("[a-zA-Z1-9]+@[a-zA-Z1-9_-]+(\\.[a-zA-Z1-9_-]+)+"))
						System.out.println("Invalid e-mail address format.  Please enter your e-mail address again.");
					else
						validInput = true;
				} catch (IOException e) {
					//e.printStackTrace();
				}
			}
			
			// Verify that the user is a member of the project.
			if (userType == 1) {
				validUser = verifyUser(user, eMail);
				if (validUser == false)
					System.out.println("The user name and e-mail address you entered was not found.  Please enter them again.");
			}
			
			// Add the user to the project.
			else if (userType == 2) {
				validUser = addUser(user, eMail);
				if (validUser == false)
					System.out.println("The user name and e-mail address you entered is already in use.  Please enter them again.");
			}
		}
		
		// Write the properties file.
		OutputStream os = null;
		try {
			os = new FileOutputStream(CONFIG_FILE_PATH);
		} catch (FileNotFoundException e) {
			SysLogger.println("ERROR: Could not open " + CONFIG_FILE_PATH + " for writing.");
			System.out.println("ERROR: Could not open " + CONFIG_FILE_PATH + " for writing.");
		}
		Properties props = new Properties();
		props.setProperty("user_name", user);
		props.setProperty("email_address", eMail);
		try {
			props.store(os, null);
		} catch (IOException e) {
			SysLogger.println("ERROR: Could not write to " + CONFIG_FILE_PATH + ".");
			System.out.println("ERROR: Could not write to " + CONFIG_FILE_PATH + ".");
		}
		
		this.userName = user;
		this.eMailAddress = eMail;
	}
	
	/**
	 * Adds a user to the project.
	 * @param userName The user name to add.
	 * @param eMailAddress The user's e-mail address.
	 * @return Returns true if the user was successfully added, false if the user already existed.
	 */
	private boolean addUser(String userName, String eMailAddress) {
		return executeUserRPC("addUser", userName, eMailAddress);
	}
	
	/**
	 * Determines whether a given user is already a member of the project.
	 * @param userName The user name to search for.
	 * @param eMailAddress The e-mail address of the user.
	 * @return Returns true if the user is a member of the project, false otherwise.
	 */
	private boolean verifyUser(String userName, String eMailAddress) {
		return executeUserRPC("verifyUser", userName, eMailAddress);
	}
	
	/**
	 * Executes one of the user XML-RPCs.
	 * @param methodName The name of the method to execute.
	 * @param userName The user name to search for.
	 * @param eMailAddress The user's e-mail address.
	 * @return The return value depends on the outcome of the RPC.
	 */
	private boolean executeUserRPC(String methodName, String userName, String eMailAddress) {
		Vector<String> args = new Vector<String>();
		args.add(userName);
		args.add(eMailAddress);
		boolean success = false;
		
		try {
			Object result = getXmlRpcPRCServerClient().execute("server." + methodName, args);
			success = ((Boolean)result).booleanValue();
		} catch (Exception e) {
			SysLogger.println("PRCClient.executeUserRPC: Error executing server." + methodName + ": " + e);
			System.out.println("ERROR: Could not connect to the project server.");
			System.err.println("ERROR: Could not connect to the project server.");
			shutdown();			
		}
		
		return success;
	}
	
	/**
	 * @return Returns the xmlrpcServer.
	 */
	public WebServer getXmlRpcServer() {
		return xmlrpcServer;
	}
	
	/**
	 * @return Returns the XML-RPC client for the project server.
	 */
	public XmlRpcClient getXmlRpcPRCServerClient() {
		return xmlrpcPRCServerClient;
	}
	
	/**
	 * @return Returns the XML-RPC client for the science application.
	 */
	public XmlRpcClient getXmlRpcScienceAppClient() {
		return xmlrpcScienceAppClient;
	}
	
	/**
	 * @return Returns the userName.
	 */
	public String getUserName() {
		return userName;
	}
	
	/**
	 * @return Returns the eMailAddress.
	 */
	public String getEMailAddress() {
		return eMailAddress;
	}
	
	/**
	 * Requests a new work unit for this user from the project server.
	 * @return The new work unit for this user, or null if the server did not have any work units.
	 */
	public NonPersistentWorkUnit getWorkUnit() {
		SysLogger.println("PRCClient.getWorkUnit: Requesting new work unit from server.");
		Vector<String> args = new Vector<String>();
		args.add(this.userName);
		args.add(this.eMailAddress);
		Vector wuVector = null;
		NonPersistentWorkUnit wu = null;
		
		Object result = null;
		try {
			result = getXmlRpcPRCServerClient().execute("server.getWorkUnit", args);
			
			try {
				// Build the work unit.
				wuVector = (Vector)result;
				wu = new NonPersistentWorkUnit();
				wu.setTaskID((String)wuVector.get(0));
				wu.setData((byte[])wuVector.get(1));
				SysLogger.println("PRCClient.getWorkUnit: Received work unit " + wu.getTaskID() + ".");
				
				// Write the work unit to disk in case the client is shut down.
				writeWorkUnitToDisk(wu.getTaskID(), wu.getData());
			} catch (Exception e) {
				try {
					Exception e1 = (Exception)result;
					SysLogger.println("PRCClient.getWorkUnit: Received an Exception from the server while requesting a new work unit: " + e1);
					SysLogger.println("PRCClient.getWorkUnit: Waiting for " + (SERVER_RETRY_DELAY / 1000 / 60) + " minutes before retry attempt.");
				} catch (Exception e2) {
					SysLogger.println("PRCClient.getWorkUnit: Received an Exception from the server while requesting a new work unit: " + e);
					SysLogger.println("PRCClient.getWorkUnit: Waiting for " + (SERVER_RETRY_DELAY / 1000 / 60) + " minutes before retry attempt.");
				}
			}
		} catch (XmlRpcException e) {
			SysLogger.println("PRCClient.getWorkUnit: Received an Exception from the server while requesting a new work unit: " + e);
		} catch (IOException e) {
			SysLogger.println("PRCClient.getWorkUnit: Received an Exception from the server while requesting a new work unit: " + e);
		} catch (Exception e) {
			SysLogger.println("PRCClient.getWorkUnit: Received an Exception from the server while requesting a new work unit: " + e);
		}
		
		return wu;
	}
	
	/**
	 * Reads the work unit files on disk if they exist.
	 * @return Returns the work unit that was read from the disk, or null if none was found.
	 */
	public NonPersistentWorkUnit readWorkUnitFromDisk() {
		SysLogger.println("PRCClient.readWorkUnitFromDisk: Reading work unit from disk.");
		String id = null;
		byte[] data = null;
		NonPersistentWorkUnit wu = null;
		InputStream is;
		
		// Read work unit ID.
		try {
			is = new FileInputStream(WORK_UNIT_ID_PATH);
			byte[] idBytes = new byte[is.available()];
			is.read(idBytes);
			id = new String(idBytes);
		} catch (FileNotFoundException e) {
			SysLogger.println("PRCClient.readWorkUnitFromDisk: Could not open file " + WORK_UNIT_ID_PATH);
		} catch (IOException e) {
			SysLogger.println("PRCClient.readWorkUnitFromDisk: Error reading from file " + WORK_UNIT_ID_PATH);
		}
		
		// Read work unit data.
		try {
			is = new FileInputStream(WORK_UNIT_DATA_PATH);
			data = new byte[is.available()];
			is.read(data);
		} catch (FileNotFoundException e) {
			SysLogger.println("PRCClient.readWorkUnitFromDisk: Could not open file " + WORK_UNIT_DATA_PATH);
		} catch (IOException e) {
			SysLogger.println("PRCClient.readWorkUnitFromDisk: Error reading from file " + WORK_UNIT_DATA_PATH);
		}
		
		if ((id != null) && (data != null)) {
			wu = new NonPersistentWorkUnit(id);
			wu.setData(data);
		}
		
		return wu;
	}
	
	/**
	 * Writes the work unit ID and data to disk.
	 * @param id The work unit ID to write.
	 * @param data The work unit data to write.
	 */
	private void writeWorkUnitToDisk(String id, byte[] data) {
		SysLogger.println("PRCClient.writeWorkUnitToDisk: Writing work unit " + id + " to disk.");
		OutputStream os;
		
		// Write work unit ID.
		try {
			os = new FileOutputStream(WORK_UNIT_ID_PATH);
			os.write(id.getBytes());
			os.flush();
			os.close();
		} catch (FileNotFoundException e) {
			SysLogger.println("PRCClient.writeWorkUnitToDisk: Could not open file " + WORK_UNIT_ID_PATH);
		} catch (IOException e) {
			SysLogger.println("PRCClient.writeWorkUnitToDisk: Error writing to file " + WORK_UNIT_ID_PATH);
		}
		
		// Write work unit data.
		try {
			os = new FileOutputStream(WORK_UNIT_DATA_PATH);
			os.write(data);
			os.flush();
			os.close();
		} catch (FileNotFoundException e) {
			SysLogger.println("PRCClient.writeWorkUnitToDisk: Could not open file " + WORK_UNIT_DATA_PATH);
		} catch (IOException e) {
			SysLogger.println("PRCClient.writeWorkUnitToDisk: Error writing to file " + WORK_UNIT_DATA_PATH);
		}
	}
	
	/**
	 * Deletes the work unit files from disk.
	 */
	public void deleteWorkFilesFromDisk() {
		File idFile = new File(WORK_UNIT_ID_PATH);
		File dataFile = new File(WORK_UNIT_DATA_PATH);
		File checkpointFile = new File(CHECKPOINT_PATH);
		
		// Delete the files.
		idFile.delete();
		dataFile.delete();
		checkpointFile.delete();
	}
	
	/**
	 * Runs the science application, then passes it a work unit to compute 
	 * and waits for the science application to return a result.
	 * @param wu The work unit for which to compute a result.
	 * @return Returns the result computed by the science application. 
	 */
	public NonPersistentResult invokeScienceApp(NonPersistentWorkUnit wu) {
		SysLogger.println("PRCClient.invokeScienceApp: Sending work unit " + wu.getTaskID() + " to science application for processing.");
		
		Vector<byte[]> args = new Vector<byte[]>();
		args.add(wu.getData());
		Vector resultVector = null;
		boolean receivedResult = false;
		
		// Execute the XML-RPC call.
		while (receivedResult == false) {
			try {
				Object result = getXmlRpcScienceAppClient().execute("sciapp.computeResult", args);
				try {
					resultVector = (Vector)result;
					receivedResult = true;
				} catch (Exception e) {
					SysLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
					SysLogger.println("PRCClient.invokeScienceApp: Waiting for " + (SCIAPP_RETRY_DELAY / 1000) + " seconds before retry attempt.");
					try {
						Thread.sleep(SCIAPP_RETRY_DELAY);
					} catch (InterruptedException e1) {
						SysLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
					}
				}
			} catch (XmlRpcException e) {
				SysLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
				SysLogger.println("PRCClient.invokeScienceApp: Waiting for " + (SCIAPP_RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(SCIAPP_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			} catch (IOException e) {
				// If control reaches this block, the science application is not running.
				SysLogger.println("PRCClient.invokeScienceApp: Science Application is not running.");
				SysLogger.println("PRCClient.invokeScienceApp: Starting Science Application.");
				try {
					Runtime.getRuntime().exec(project.getScienceAppPath(), null, new File(project.getClientDirectoryPath()));
				} catch (IOException e1) {
					SysLogger.println("PRCClient.invokeScienceApp: Error starting Science Application: " + e1);
					e1.printStackTrace();
				}
				
				// Wait for the science application to start up before sending any work units.
				try {
					Thread.sleep(STARTUP_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			} catch (Exception e) {
				SysLogger.println("PRCClient.invokeScienceApp: Received an Exception from the science application while computing a result: " + e);
				SysLogger.println("PRCClient.invokeScienceApp: Waiting for " + (SCIAPP_RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(SCIAPP_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("PRCClient.invokeScienceApp: Exception thrown while sleeping: " + e1);
				}
			}
		}
		
		// Build the result.
		NonPersistentResult r = new NonPersistentResult();
		r.setData((byte[])resultVector.get(0));
		
		SysLogger.println("PRCClient.invokeScienceApp: Received result from science application for work unit " + wu.getTaskID() + ".");
		
		return r;
	}
	
	/**
	 * Returns a result to the server.
	 * @param wu The WorkUnit for which the result was computed.
	 * @param r The result that was computed.
	 * @return Returns true if the result was successfully sent to the server and acknowledged by the server, false otherwise.
	 */
	@SuppressWarnings("unchecked")
	public boolean returnResult(NonPersistentWorkUnit wu, NonPersistentResult r) {
		SysLogger.println("PRCClient.returnResult: Returning result for work unit " + wu.getTaskID() + ".");
		boolean success = false;
		boolean serverSuccess = false;
		
		// Construct the result vector.
		Vector returnValue = new Vector();
		returnValue.add(userName);
		returnValue.add(eMailAddress);
		returnValue.add(wu.getTaskID());
		returnValue.add(r.getData());
		
		// Send the result vector to the server.
		Object serverReturn = null;
		while (success == false) {
			try {
				serverReturn = getXmlRpcPRCServerClient().execute("server.returnResult", returnValue);
				success = true;
				serverSuccess = (Boolean)serverReturn;
			} catch (ClassCastException e) {
				success = true;
				serverSuccess = false;
				try {
					Exception e1 = (Exception)serverReturn;
					SysLogger.println("PRCClient.returnResult: Server threw exception: " + e1.getCause());
				} catch (Exception e2) {
					SysLogger.println("PRCClient.returnResult: Server threw exception: " + e);
				}
			} catch (XmlRpcException e) {
				SysLogger.println("Error sending result to server: " + e);
				try {
					Thread.sleep(SERVER_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("PRCClient.returnResult: Exception thrown while sleeping: " + e1);
				}
			} catch (IOException e) {
				SysLogger.println("Error sending result to server: " + e);
				try {
					Thread.sleep(SERVER_RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("PRCClient.returnResult: Exception thrown while sleeping: " + e1);
				}
			}
		}
		
		return serverSuccess;
	}

	/**
	 * @return Returns true if the server is running, false otherwise.
	 */
	public boolean isRunning() {
		return isRunning;
	}
	
	/**
	 * Starts the XML-RPC server and begins handling client requests.
	 */
	public void run() {
		if (isRunning == false) {
			SysLogger.println("PRC Client Starting XML-RPC Server.");
			System.out.println("PRC Client Starting XML-RPC Server.");
			xmlrpcServer.start();
			isRunning = true;
		}
	}

	/**
	 * Performs any tasks necessary to clearnly shut down the client.
	 */
	public void shutdown() {
		// Shut down the science application.
		SysLogger.println("Shutting down PRC Client.");
		System.out.println("Shutting down PRC Client.");
		try {
			xmlrpcScienceAppClient.execute("sciapp.shutdown", new Vector());
		} catch (XmlRpcException e) {
			SysLogger.println("Error shutting down science application: " + e);
		} catch (IOException e) {
			SysLogger.println("Error shutting down science application: " + e);
		}
		
		isRunning = false;
		clientThread.interrupt();
	}

	/**
	 * @return Returns the singleton instance of the PRCClient class.
	 * @throws MissingConfigurationFileException
	 */
	public static PRCClient instance() throws MissingConfigurationFileException {
		if (prcClient == null)
			prcClient = new PRCClient();
		return prcClient;
	}
	
	/**
	 * Controls the client.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		try { 
			SysLogger.initialize(LOG_PATH);
		} catch (InvalidLogPathException e) {
			e.printStackTrace();
		}
		PRCClient client = null;
		try {
			client = PRCClient.instance();
		} catch (MissingConfigurationFileException e) {
			System.out.println("Error: Project configuration file not found.  Exiting.");
			System.err.println("Error: Project configuration file not found.  Exiting.");
			System.exit(EXIT_ERROR);
		}
		
		client.run();
		System.out.println("PRC Client Ready.");
		
		// First try to read the work unit from disk.
		NonPersistentWorkUnit wu = client.readWorkUnitFromDisk();
		
		// If there is no work unit on disk, get a work unit from the server.
		if (wu == null)
			wu = client.getWorkUnit();
		
		while (client.isRunning()) {
			if (wu != null) {
				// Compute the result.
				NonPersistentResult r = client.invokeScienceApp(wu);
				
				// Send the result to the server.
				boolean serverReturn = client.returnResult(wu, r);
				if (serverReturn == true)
					SysLogger.println("Server successfully received and acknowledged the result for work unit " + wu.getTaskID() + ".");
				else
					SysLogger.println("Server successfully received but did not acknowledge the result for work unit " + wu.getTaskID() + ".");

				// The client is done with this work unit, so the work files can be deleted from disk.
				client.deleteWorkFilesFromDisk();
			}
			else {
				try {
					Thread.sleep(SERVER_RETRY_DELAY);
				} catch (InterruptedException e) {
					// This is expected if the client.shutdown RPC was executed.
				}
			}
			
			if (client.isRunning()) {
				// Get the next work unit
				wu = client.getWorkUnit();
			}
		}
		
		// Shut down the XML-RPC server.
		client.getXmlRpcServer().shutdown();
		
		System.out.println("PRC Client Halted.");
		System.exit(EXIT_SUCCESS);
	}
}
