package edu.wpi.prc.client;

import java.io.IOException;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

/**
 * The XMLRPCClient allows a project administrator to send RPCs to
 * any XML-RPC server from the command line.
 * @author James Baldassari
 */
public class XMLRPCClient {
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private static XmlRpcClient xmlrpcClient = null;
	
	/**
	 * Parses command-line arguments and sends XML-RPCs to the server.
	 * @param args The command-line arguments.
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		if (args.length < 2)
			printUsageAndExit();
		
		// For some reason the Apache XML-RPC library sends output to System.err.
		// It's really annoying, so just redirect System.err into oblivion.
		System.setErr(new PrintStream(new PipedOutputStream()));
		
		// Instantiate an XML-RPC client.
		try {
			xmlrpcClient = new XmlRpcClient(args[0]);
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + args[0] + "\" is invalid.");
			System.exit(EXIT_ERROR);
		}
		
		// Build Vector of parameters.
		Vector params = new Vector();
		for (int ii = 2; ii < args.length; ii++)
			params.add(args[ii]);
		
		// Send the RPC.
		if (sendRPC(args[1], params) == false)
			System.exit(EXIT_ERROR);
		else
			System.exit(EXIT_SUCCESS);
	}

	/**
	 * Executes the given XML-RPC command with the given parameters. 
	 * @param cmd The command to send.
	 * @param params The parameters to pass.
	 * @return True if the command was sent successfully, false otherwise.
	 */
	private static boolean sendRPC(String cmd, Vector params) {
		System.out.println("Sending: " + params);
		try {
			Object result = xmlrpcClient.execute(cmd, params);
			try {
				Vector v = (Vector)result;
				System.out.println("Server returned vector:");
				for (int ii = 0; ii < v.size(); ii++) {
					try {
						String element = (String)v.get(ii);
						System.out.println("\t" + element);
						return true;
					} catch (ClassCastException e) {
						byte[] bytes = (byte[])v.get(ii);
						System.out.print("\tByte Array: ");
						for (byte b : bytes)
							System.out.print(b + " ");
						System.out.println();
					}
				}
				return true;
			} catch (ClassCastException e) {
				System.out.println("Server returned: " + result);
				return true;
			}
		} catch (XmlRpcException e) {
			//System.out.println("Error sending command: " + e);
			return false;
		} catch (IOException e) {
			//System.out.println("Error sending command: " + e);
			return false;
		}
	}
	
	/**
	 * Prints the usage, and exits the program.
	 */
	private static void printUsageAndExit() {
		System.out.println("Usage: XMLRPCClient server_URL command [arguments...]");
		System.out.println("Example: XMLRPCClient http://localhost:2080/RPC2 server.shutdown");
		System.out.println("Example: XMLRPCClient http://localhost:2080/RPC2 server.getWorkUnit jdb");
		System.exit(EXIT_ERROR);
	}
}
