/**
 * 
 */
package edu.wpi.prc.persistence;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;


/**
 * Logs failed spot checks to the spot check log file.
 * @author James Baldassari
 */
public class SpotCheckLogger {
	private static SpotCheckLogger logger = null;
	public static final String DEFAULT_LOG_DIR = "log";
	public static final String DEFAULT_LOG_PATH = DEFAULT_LOG_DIR + File.separator + "spotcheck.log";
	private static final String LOG_TIMESTAMP_FORMAT = new String("EEE MMM dd HH:mm:ss:SSS z yyyy");
	private static SimpleDateFormat timestampFormat = new SimpleDateFormat(LOG_TIMESTAMP_FORMAT);
	private File logFile;
	private FileOutputStream logFileStream;
	private PrintStream logStream;
	
	/**
	 * Private constructor.
	 */
	private SpotCheckLogger() {
		
	}
	
	/**
	 * Opens the default log file for appending and initializes the streams.
	 * @throws InvalidLogPathException
	 */
	public static void initialize() throws InvalidLogPathException {
		initialize(null);
	}
	
	/**
	 * Opens the log file for appending and initializes the streams.
	 * @param logPath The path to the log file.
	 * @throws InvalidLogPathException
	 */
	public static void initialize(String logPath) throws InvalidLogPathException {
		// Instantiate the logger if necessary.
		if (logger == null)
			logger = new SpotCheckLogger();
		
		// Set default log path if logPath is null.
		if (logPath != null)
			logger.logFile = new File(logPath);
		else
			logger.logFile = new File(DEFAULT_LOG_PATH);

		// Try to open the given log path for appending.
		try {
			logger.logFileStream = new FileOutputStream(logger.logFile, true);
			logger.logStream = new PrintStream(logger.logFileStream);

		} catch (FileNotFoundException e) {
			System.out.println("Error: Could not open file \"" + logPath + "\" for appending.");
			System.err.println("Error: Could not open file \"" + logPath + "\" for appending.");
			
			// If an error occurred while opening the default path, throw an exception.
			if (logPath == null)
				throw new InvalidLogPathException("Could not open log file for appending.");
			
			// Otherwise, fail over to the default path.
			else {
				// Create the log directory if it doesn't exist.
				File logDir = new File("log");
				if (logDir.exists() == false) {
					if (logDir.mkdir() == false)
						throw new InvalidLogPathException("Could not create \"log\" directory.");
				}
				else if (logDir.isDirectory() == false)
					throw new InvalidLogPathException("\"log\" exists but is not a directory.");
				else {
					logger.logFile = new File(DEFAULT_LOG_PATH);
					try {
						if (logger.logFile.exists() && (logger.logFile.canWrite() == false))
							throw new InvalidLogPathException("No write access for default log file: " + DEFAULT_LOG_PATH + ".");
						else {
							logger.logFileStream = new FileOutputStream(logger.logFile, true);
							logger.logStream = new PrintStream(logger.logFileStream);
						}

					} catch (FileNotFoundException e2) {
						System.out.println("Error: Could not open default log file (" + DEFAULT_LOG_PATH + ") for appending.");
						System.err.println("Error: Could not open default log file (" + DEFAULT_LOG_PATH + ") for appending.");
						throw new InvalidLogPathException("Could not open default log file for appending.");
					}
				}
			}
		}
	}
	
	/**
	 * Writes the given string to the log file.
	 * NOTE: initialize() must be called first or print() will have no effect.
	 * @param str The string to write.
	 */
	public static void print(String str) {
		if (logger != null)
			logger.logStream.print(timestampFormat.format(new Date()) + ": " + str);
	}
	
	/**
	 * Writes the given string to the log file, then writes a new line.
	 * NOTE: initialize() must be called first or println() will have no effect.
	 * @param str The string to write.
	 */
	public static void println(String str) {
		if (logger != null)
			logger.logStream.println(timestampFormat.format(new Date()) + ": " + str);
	}
}
