package edu.wpi.prc.server;

import java.io.File;
import java.net.MalformedURLException;
import java.util.concurrent.Semaphore;

import org.apache.xmlrpc.WebServer;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.persistence.InvalidLogPathException;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.server.transitioner.LocalTransitioner;
import edu.wpi.prc.server.transitioner.RemoteTransitionerProxy;
import edu.wpi.prc.server.transitioner.Transitioner;
import edu.wpi.prc.server.validator.DefaultValidatorThread;

public class PRCServer implements Runnable {
	private static PRCServer prcServer = null;
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private Transitioner transitioner = null;
	private final ServerRPCHandler rpcHandler = new ServerRPCHandler();
	private WebServer xmlrpcServer = null;
	private static boolean isRunning = false;
	private static boolean isTransitionerRunning = false;
	private static boolean waitForTransactions = true;	// Wait for transitioner transactions to complete before shutting down?
	private static final Semaphore transitionerSemaphore = new Semaphore(1, true);
	
	/**
	 * Private constructor prevents explcit instantiation of the PRCServer class.
	 * Initializes the server componenets.
	 */
	private PRCServer() {			
		transitioner = getTransitioner();
		initXmlRpcServer();
		SysLogger.println("PRC server initialized.");
		System.out.println("PRC server initialized.");
	}
	
	/**
	 * Initializes and starts the XML-RPC server.
	 */
	private void initXmlRpcServer() {
		// Get the server URI from the Project class.
		String serverURI = Project.instance().getServerURI();
		
		// Make sure that the server URI has been configured before proceeding.
		if (serverURI == null) {
			// If it was not in the database, try reading from the configuration file.
			Project p = null;
			try {
				Project.readFromFile();
			} catch (MissingConfigurationFileException e) {
				
			}
			if (p.getServerURI() != null)
				p.saveOrUpdate();
			else {
				System.err.println("Fatal error: The server URI is not configured.");
				System.exit(EXIT_ERROR);
			}
		}
		
		// Initialize the server.
		try {
			// Parse the server port.
			String serverPortStr = serverURI.split(":")[1];
			int serverPort = Integer.parseInt(serverPortStr);
			
			// Instantiate the web server.
			xmlrpcServer = new WebServer(serverPort);
			
			// Register RPC handler.
			xmlrpcServer.addHandler("server", rpcHandler);
		} catch (Exception e) {
			SysLogger.println("Error initializing the XML-RPC server: " + e);
			System.err.println("Error initializing the XML-RPC server: " + e);
			System.exit(EXIT_ERROR);
		}
	}
	
	/**
	 * @return Returns the transitioner used by this project.
	 */
	public Transitioner getTransitioner() {
		if (transitioner == null) {
			String transitionerURI = Project.instance().getTransitionerURI();
			if (transitionerURI == null)
				transitioner = new LocalTransitioner();
			else {
				try {
					transitioner = new RemoteTransitionerProxy("http://" + transitionerURI + "/RPC2");
				} catch (MalformedURLException e) {
					SysLogger.println("ERROR: Invalid URI \"" + transitionerURI + "\" Specified for Remote Transitioner.");
					System.out.println("ERROR: Invalid URI \"" + transitionerURI + "\" Specified for Remote Transitioner.");
					SysLogger.println("Defaulting to Local Transitioner.");
					System.out.println("Defaulting to Local Transitioner.");
					transitioner = new LocalTransitioner();
				}
			}
			
			isTransitionerRunning = true;
		}
		
		return transitioner;
	}

	/**
	 * @param waitForTransactions True if the server should wait for all transitioner transactions to complete before shutting down, false otherwise.
	 */
	public void setWaitForTransactions(boolean waitForTransactions) {
		PRCServer.waitForTransactions = waitForTransactions;
	}

	/**
	 * @return Returns the transitioner semaphore.
	 */
	public Semaphore getTransitionerSemaphore() {
		return transitionerSemaphore;
	}
	
	/**
	 * @return Returns the RPC handler.
	 */
	public ServerRPCHandler getRPCHandler() {
		return rpcHandler;
	}
	
	/**
	 * @return Returns the XML-RPC server.
	 */
	public WebServer getXmlrpcServer() {
		return xmlrpcServer;
	}
	
	/**
	 * @return Returns true if the server is running, false otherwise.
	 */
	public boolean isRunning() {
		return isRunning;
	}
	
	/**
	 * @return Returns true if the transitioner is running, false otherwise.
	 */
	public boolean isTransitionerRunning() {
		return isTransitionerRunning;
	}
	
	/**
	 * Starts the XML-RPC server and begins handling client requests.
	 */
	public void run() {
		if (isRunning == false) {
			// Start the XML-RPC server thread.
			SysLogger.println("Starting XML-RPC Server.");
			System.out.println("Starting XML-RPC Server.");
			xmlrpcServer.start();
			
			// If there is no validator defined for the project, start a default result validator thread.
			if (Project.instance().getDefaultValidator() == true) {
				SysLogger.println("Starting Default Validator Thread.");
				System.out.println("Starting Default Validator Thread.");				
				DefaultValidatorThread resultValidator = new DefaultValidatorThread();
				resultValidator.start();
			}
				
			isRunning = true;
		}
	}
	
	/**
	 * Switches the current transitioner to a LocalTransitioner.
	 * NOTE: This functionality is not fully supported and is intended to be a future feature.
	 */
	public void switchToLocalTransitioner() {
		SysLogger.println("Switching to Local Transitioner.");
		System.out.println("Switching to Local Transitioner.");
		try {
			shutdownTransitioner(true);
			transitioner = new LocalTransitioner();
		} finally {
			isTransitionerRunning = true;
			transitionerSemaphore.release();
		}
		SysLogger.println("Now Using a Local Transitioner.");
		System.out.println("Now Using a Local Transitioner.");
	}
	
	/**
	 * Switches the current transitioner to a RemoteTransitioner.
	 * NOTE: This functionality is not fully supported and is intended to be a future feature.
	 */
	public void switchToRemoteTransitioner() {
		SysLogger.println("Switching to Remote Transitioner.");
		System.out.println("Switching to Remote Transitioner.");
		String transitionerURI = Project.instance().getTransitionerURI(); 
		if (transitionerURI != null) {
			try {
				shutdownTransitioner(true);
				transitioner = new RemoteTransitionerProxy("http://" + transitionerURI + "/RPC2");
				SysLogger.println("Now Using a Remote Transitioner.");
				System.out.println("Now Using a Remote Transitioner.");
			} catch (MalformedURLException e) {
				SysLogger.println("ERROR: Invalid URI \"" + transitionerURI + "\" Specified for Remote Transitioner.");
				System.out.println("ERROR: Invalid URI \"" + transitionerURI + "\" Specified for Remote Transitioner.");
				SysLogger.println("Defaulting to Local Transitioner.");
				System.out.println("Defaulting to Local Transitioner.");
				transitioner = new LocalTransitioner();
				SysLogger.println("Now Using a Local Transitioner.");
				System.out.println("Now Using a Local Transitioner.");
			} finally {
				isTransitionerRunning = true;
				transitionerSemaphore.release();
			}
		}
	}

	/**
	 * Performs any tasks necessary to clearnly shut down the server.
	 */
	public void initiateShutdown() {
		// Shut down the XML-RPC server.
		SysLogger.println("Shutting down PRC Server.");
		System.out.println("Shutting down PRC Server.");
		isRunning = false;
	}
	
	/**
	 * Cleanly shuts down the transitioner.
	 */
	public void shutdownTransitioner(boolean waitForTransactions) {
		// Acquire lock on the transitioner so that no new 
		// XML-RPCs can access the transitioner.
		try {
			transitionerSemaphore.acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		isTransitionerRunning = false;
		
		// Shut down transitioner.
		SysLogger.println("Shutting down Transitioner.");
		System.out.println("Shutting down Transitioner.");
		
		// Wait for all transitioner transactions to complete unless this method was called from an emergency shutdown command.
		if (PRCServer.waitForTransactions == true) {
			if (rpcHandler.getNumTransactions() > 0) {
				SysLogger.println("Waiting for outstanding transitioner transactions to complete.  This may take a few minutes.");
				System.out.println("Waiting for outstanding transitioner transactions to complete.  This may take a few minutes.");			
				while ((PRCServer.waitForTransactions == true) && (rpcHandler.getNumTransactions() > 0)) {
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						
					}
				}
				if (PRCServer.waitForTransactions ==  false) {
					SysLogger.println("WARNING: Shutting down transitioner without waiting for outstanding transactions to complete.  The database may be inconsistent.");
					System.out.println("WARNING: Shutting down transitioner without waiting for outstanding transactions to complete.  The database may be inconsistent.");
				}
			}
		}
		else {
			SysLogger.println("WARNING: Shutting down transitioner without waiting for outstanding transactions to complete.  The database may be inconsistent.");
			System.out.println("WARNING: Shutting down transitioner without waiting for outstanding transactions to complete.  The database may be inconsistent.");
		}
		
		// It is now safe to shut down the transitioner.
		getTransitioner().shutdown();
		SysLogger.println("Transitioner Halted.");
		System.out.println("Transitioner Halted.");
	}
	
	/**
	 * Shuts down the server when it is safe to do so.
	 * This method should never be called directly.
	 * Instead, initiateShutdown() should be called, 
	 * which will cause the main() to call this method.
	 */
	public void shutdown() {
		shutdownTransitioner(waitForTransactions);
		
		// Shut down XML-RPC server.
		SysLogger.println("Shutting down XML-RPC Server.");
		System.out.println("Shutting down XML-RPC Server.");
		getXmlrpcServer().shutdown();
		SysLogger.println("XML-RPC Server Halted.");
		System.out.println("XML-RPC Server Halted.");
		
		// Shut down Hibernate.
		SysLogger.println("Shutting down Hibernate.");
		System.out.println("Shutting down Hibernate.");
		HibernateController.instance().shutdown();
		SysLogger.println("Hibernate Halted.");
		System.out.println("Hibernate Halted.");
		
		SysLogger.println("PRC Server Halted.");
		System.out.println("PRC Server Halted.");
	}
	
	/**
	 * @return Returns the singleton instance of the PRCServer class.
	 */
	public static PRCServer instance() {
		if (prcServer == null)
			prcServer = new PRCServer();
		return prcServer;
	}
	
	/**
	 * Controls the server.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		// Try to initialize Hibernate.
		// If the project configuration file does not exist, the server will exit immediately.
		try {
			HibernateController.instance().initialize();
		} catch (MissingConfigurationFileException e) {
			System.out.println(e);
			System.err.println(e);
			System.exit(EXIT_ERROR);
		}
		
		// Now we can use Project.instance() to get the rest of the project configuration.
		try { 
			SysLogger.initialize("log" + File.separator + "server.log");
		} catch (InvalidLogPathException e) {
			e.printStackTrace();
		}
		PRCServer server = PRCServer.instance();
		
		// Start the server.
		server.run();
		
		SysLogger.println("PRC Server Ready.");
		System.out.println("PRC Server Ready.");
		
		// XML-RPC server is started in a new thread, so just 
		while (server.isRunning()) {
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		server.shutdown();
		System.exit(EXIT_SUCCESS);
	}
}
