package edu.wpi.prc.server.transitioner;

import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.persistence.SpotCheckLogger;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.project.SpotCheckFailureAction;
import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.server.spotcheck.SpotCheckThread;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.MultiPriorityQueue;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;
import edu.wpi.prc.user.VolunteerState;

/**
 * The LocalTransitioner is used when the Transitioner component is executed 
 * on the same machine as the PRCServer.  If the Transitioner will be run on 
 * a separate machine, use the RemoteTransitionerProxy instead.
 * @see Transitioner
 * @see RemoteTransitionerProxy
 * @author James Baldassari
 */
public class LocalTransitioner implements Transitioner {
	private TransitionerQueue<WorkUnit> ingressWorkUnits = new MultiPriorityQueue<WorkUnit>();
	private TransitionerQueue<WorkUnit> pendingWorkUnits = new MultiPriorityQueue<WorkUnit>();
	private TransitionerQueue<WorkUnit> spotCheckWorkUnits = new MultiPriorityQueue<WorkUnit>();
	private TransitionerQueue<WorkUnit> canonicalWorkUnits = new MultiPriorityQueue<WorkUnit>();	// WUs for which a canonical result can be chosen.
	private TransitionerQueue<Result> ingressResults = new MultiPriorityQueue<Result>();
	private TransitionerQueue<Result> pendingResults = new MultiPriorityQueue<Result>();
	private TransitionerQueue<Result> spotCheckResults = new MultiPriorityQueue<Result>();
	private SpotCheckThread spotChecker = null;
	private final int minNumResults = Project.instance().getMinNumResults();
	private final float spotCheckProbability = Project.instance().getSpotCheckProbability();
	private final int minNumSpotChecks = Project.instance().getMinNumSpotChecks(); 
	
	/**
	 * Private constructor ensures that only one instance of LocalTransitioner exists.
	 * Use the getTransitioner() method to get access to the LocalTransitioner.
	 */
	public LocalTransitioner() {	
		restoreFromDatabase();
		verifySpotChecks();
	}
	
	/**
	 * @return Returns the ingress work queue. 
	 */
	public TransitionerQueue<WorkUnit> getIngressWorkQueue() {
		return ingressWorkUnits;
	}
	
	/**
	 * @return Returns the pending work queue.
	 */
	public TransitionerQueue<WorkUnit> getPendingWorkQueue() {
		return pendingWorkUnits;
	}
	
	/**
	 * @return Returns the work units that have enough results to select a canonical result.
	 */
	public TransitionerQueue<WorkUnit> getCanonicalWorkQueue() {
		return canonicalWorkUnits;
	}
	
	/**
	 * @return Returns the retired work queue.
	 */
	public TransitionerQueue<WorkUnit> getRetiredWorkQueue() {
		WorkUnit[] retiredWorkUnits = WorkUnit.findAllEgress();
		TransitionerQueue<WorkUnit> retiredQueue = new MultiPriorityQueue<WorkUnit>();
		retiredQueue.addAll(retiredWorkUnits);
		return retiredQueue;
	}
	
	/**
	 * @return Returns the ingress result queue.
	 */
	public TransitionerQueue<Result> getIngressResultQueue() {
		return ingressResults;
	}
	
	/**
	 * @return Returns the pending result queue.
	 */
	public TransitionerQueue<Result> getPendingResultQueue() {
		return pendingResults;
	}
	
	/**
	 * @return Returns the valid result queue, consisting of both valid results and accepted spot-check results.
	 */
	public TransitionerQueue<Result> getValidResultQueue() {
		Result[] validResults = Result.findAllEgress();
		Result[] spotCheckResults = Result.findAllSpotCheck();
		TransitionerQueue<Result> validResultQueue = new MultiPriorityQueue<Result>();
		validResultQueue.addAll(validResults);
		validResultQueue.addAll(spotCheckResults);
		return validResultQueue;
	}
	
	/**
	 * Moves a WorkUnit from a work queue to the pending work queue.
	 * @param originQueue The queue from which the WorkUnit will be moved.
	 * @param wu The WorkUnit to move.
	 */
	private void moveToPendingWorkQueue(TransitionerQueue<WorkUnit> originQueue, WorkUnit wu) {	
		originQueue.remove(wu);
		wu.setState(TaskState.Pending);
		wu.update();
		pendingWorkUnits.add(wu);
	}
	
	/**
	 * Moves a WorkUnit from the pending work queue to the retired work queue.
	 * This actually just sets the work unit's state to Egress and updates it in the database.
	 * Retired work units are not stored in memory.
	 * @param wu The WorkUnit to move.
	 */
	private void moveToRetiredWorkQueue(WorkUnit wu) {
		pendingWorkUnits.remove(wu);
		wu.setState(TaskState.Egress);
		wu.update();
	}
	
	/**
	 * Moves a Result from a result queue to the pending result queue.
	 * @param originQueue The queue from which the Result will be moved.
	 * @param r The Result to move.
	 */
	private void moveToPendingResultQueue(TransitionerQueue<Result> originQueue, Result r) {
		originQueue.remove(r);
		r.setState(TaskState.Pending);
		r.update();
		pendingResults.add(r);
	}
	
	/**
	 * Moves a Result from the pending result queue to the valid result queue.
	 * This actually just sets the result's state to Egress and updates it in the database.
	 * Valid results are not stored in memory.
	 * @param r The Result to move.
	 */
	private void moveToValidResultQueue(Result r) {
		pendingResults.remove(r);
		r.setState(TaskState.Egress);
		r.update();
	}
	
	/**
	 * Enqueues a work unit in the ingress work queue.
	 * NOTE: wu.getState MUST be set to TaskState.Ingress before calling this method.
	 * @param wu The work unit to add.
	 */
	public synchronized void addIngressWorkUnit(WorkUnit wu) {
		assert(wu.getState() == TaskState.Ingress);
		ingressWorkUnits.add(wu);
	}
	
	/**
	 * NOTE: This method is for testing only!  It should never be used.
	 * @param wu The work unit to add.
	 */
	public void addPendingWorkUnit(WorkUnit wu) {
		wu.setState(TaskState.Pending);
		wu.update();
		pendingWorkUnits.add(wu);
	}
	
	/**
	 * Enqueues a result in the ingress result queue.
	 * NOTE: r.getState MUST be set to TaskState.Ingress before calling this method.
	 * @param r The result to add.
	 */
	public synchronized void addIngressResult(Result r) {
		assert(r.getState() == TaskState.Ingress);
		
		// Increase the user's score and reset the user's result timestamp.
		r.getVolunteer().increaseScore(r.getWorkUnit().getPoints());
		r.getVolunteer().setLastResultDate(new Date());
		r.getVolunteer().update();
		
		// If the associated WorkUnit has already been retired, just delete this superfluous Result.
		if (r.getWorkUnit().getState() == TaskState.Egress)
			r.delete();
		
		// If the WorkUnit has not been retired, process the Result.
		else
			ingressResults.add(r);
	}
	
	/**
	 * NOTE: This method is for testing only!  It should never be used.
	 * @param r The result to add.
	 */
	public void addPendingResult(Result r) {
		r.setState(TaskState.Pending);
		r.update();
		pendingResults.add(r);
	}
	
	/**
	 * Gets all results returned by the specified user.
	 * @param v The user to search for.
	 * @return An array of all matching results or an array of lenght zero if none were found.
	 */
	private Result[] getResultsFromUser(Volunteer v) {
		// Build the HQL query to find the results returned by the given user.
		String queryString = "from Result as result";								// Get Results
		queryString += " where result.user.userID='" + v.getUserID() + "'";			// with user = v
		
		return getAssociatedResults(queryString);
	}
	
	/**
	 * Marks a Result as valid.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit, 
	 * the Result is moved from the ingress result queue to 
	 * the pending result queue.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit has 
	 * been reached, all Results for that WorkUnit are removed 
	 * from the pending result queue, a canonical result is chosen, 
	 * and the canonical result is moved to the valid result queue. 
	 * @param validResult The result from the ingress result queue to be marked valid.
	 */
	public synchronized void markResultValid(Result validResult) {
		// Verify that the result exists in the ingress queue.
		if (ingressResults.contains(validResult)) {
			WorkUnit wu = validResult.getWorkUnit();
			
			// If this is a spot check result, just log the success and delete it.
			if (wu.getState() == TaskState.SpotCheck) {
				SpotCheckLogger.println("Volunteer " + validResult.getVolunteer().getUserID() + " passed a spot check.");
				ingressResults.remove(validResult);
				validResult.getWorkUnit().removeUser(validResult.getVolunteer());
				validResult.delete();
			}
			
			// Otherwise, process the result as normal
			else {
				wu.incrementNumValidResults();
				wu.update();
				
				// Move r to the pending queue
				moveToPendingResultQueue(ingressResults, validResult);
				
				/*
				 * If there are enough valid results for the associated work unit and 
				 * this work unit is not already in the canonical queue, mark the 
				 * work unit as canonical, meaning it is now possible to select a 
				 * canonical result for this work unit.
				 */
				if ((wu.getNumValidResults() >= minNumResults) && (canonicalWorkUnits.contains(validResult.getWorkUnit()) == false)) {
					canonicalWorkUnits.add(validResult.getWorkUnit());
				}
			}
		}
	}
	
	/**
	 * Marks a result as invalid.  This will remove the result 
	 * from the ingress queue and delete it from the database.  
	 * If the result is a spot check result, additional steps 
	 * may be taken depending on the project's spot check 
	 * failure action.
	 * @param invalidResult The result to be marked invalid.
	 */
	public synchronized void markResultInvalid(Result invalidResult) {
		// If this is a failed spot check, take the appropriate action.
		if (invalidResult.getWorkUnit().getState() == TaskState.SpotCheck) {
			// Always log the failure.
			SpotCheckLogger.println("Volunteer " + invalidResult.getVolunteer().getUserID() + " failed the spot check.");
			System.out.println("Volunteer " + invalidResult.getVolunteer().getUserID() + " failed the spot check.");
			
			// Ban user if configured to do so.
			if (Project.instance().getSpotCheckFailureAction() == SpotCheckFailureAction.BanUser) {
				SpotCheckLogger.println("Banning volunteer " + invalidResult.getVolunteer().getUserID() + " for failing the spot check.");
				invalidResult.getVolunteer().setState(VolunteerState.Banned);
				invalidResult.getVolunteer().update();
			}
			
			// Purge all of the user's past results if configured to do so.
			else if (Project.instance().getSpotCheckFailureAction() == SpotCheckFailureAction.PurgeResults) {
				Result[] userResults = getResultsFromUser(invalidResult.getVolunteer());
				for (Result flaggedResult : userResults) {
					/*
					 * If the result has already been accepted as valid, we need to 
					 * delete it and un-retire its work unit so that the correct 
					 * result can be computed.
					 */
					if (flaggedResult.getState() == TaskState.Egress) {
						WorkUnit retiredWU = flaggedResult.getWorkUnit();
						retiredWU.removeAllusers();
						retiredWU.setState(TaskState.Ingress);
						retiredWU.update();
						ingressWorkUnits.add(retiredWU);
					}
					
					// Otherwise, the result can just be deleted.
					else
						flaggedResult.delete();
				}
			}
		}
		
		// Remove the invalid result from the ingress queue.
		ingressResults.remove(invalidResult);
		
		// Decrease the user's score by the result's point value.
		invalidResult.getVolunteer().decreaseScore(invalidResult.getWorkUnit().getPoints());
		invalidResult.getVolunteer().update();
		
		// Remove the invalid result from the database.
		invalidResult.delete();
	}
	
	/**
	 * NOTE: This method is for testing only!  It should never be used.
	 * @param wu The work unit to add.
	 */
	public void addEgressWorkUnit(WorkUnit wu) {
		wu.setState(TaskState.Egress);
		wu.update();		
	}
	
	/**
	 * NOTE: This method is for testing only!  It should never be used.
	 * @param r The result to add.
	 */
	public void addEgressResult(Result r) {
		r.setState(TaskState.Egress);
		r.update();		
	}
	
	/**
	 * Finds all Results generated by the given HQL query.
	 * @param hqlQuery The HQL query to execute.
	 * @return An array of all Results generated by the HQL query or an array of length 0 if none were found.
	 */
	private Result[] getAssociatedResults(String hqlQuery) {
		Result[] associatedResults = null;
		
		// Execute the query.
		Session hibernateSession = HibernateController.instance().createSession();		
		Query resultQuery = hibernateSession.createQuery(hqlQuery);
		List results = resultQuery.list();
		
		associatedResults = new Result[results.size()];
		
		if (results.size() > 0) {
			// Convert the reults into a Result array.
			for (int ii = 0; ii < results.size(); ii++)
				associatedResults[ii] = (Result)results.get(ii);
		}
		
		try {
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			
		}
		
		return associatedResults;
	}
	
	/**
	 * Finds all ingress Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated ingress Results.
	 * @return Array of ingress Results associated with the given WorkUnit or an array of length 0 if none were found.
	 */
	public Result[] getAssociatedIngressResults(WorkUnit wu) {
		// Build the HQL query to find the associated ingress results.
		String queryString = "from Result as result";								// Get Results
		queryString += " where result.workUnit.taskID='" + wu.getTaskID() + "'";	// with workUnit = wu
		queryString += " and result.stateStr='" + TaskState.Ingress + "'";			// and task state = ingress

		return getAssociatedResults(queryString);
	}
	
	/**
	 * Finds all pending Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated pending Results.
	 * @return Array of pending Results associated with the given WorkUnit or an array of length 0 if none were found.
	 */
	public Result[] getAssociatedPendingResults(WorkUnit wu) {
		// Build the HQL query to find the associated ingress results.
		String queryString = "from Result as result";								// Get Results
		queryString += " where result.workUnit.taskID='" + wu.getTaskID() + "'";	// with workUnit = wu
		queryString += " and result.stateStr='" + TaskState.Pending + "'";			// and task state = pending
		
		return getAssociatedResults(queryString);
	}
	
	/**
	 * Finds all pending WorkUnits for which enough valid Results have been returned.
	 * @return Array of ingress Results associated with the given WorkUnit or an array of length 0 if none were found.
	 */
	private WorkUnit[] getWorkUnitsWithEnoughResults() {
		// Build the HQL query.
		String queryString = "from WorkUnit as wu";									// Get WorkUnits
		queryString += " where wu.numValidResults>=" + minNumResults;				// with at least minNumResults results.
		queryString += " and wu.stateStr='" + TaskState.Pending + "'";				// and task state = pending.

		return getAssociatedWorkUnits(queryString);
	}

	/**
	 * Finds all WorkUnits generated by the given HQL query.
	 * @param hqlQuery The HQL query to execute.
	 * @return An array of all WorkUnits generated by the HQL query or an array of length 0 if none were found.
	 */
	private WorkUnit[] getAssociatedWorkUnits(String hqlQuery) {
		WorkUnit[] associatedWorkUnits;
		
		// Execute the query.
		Session hibernateSession = HibernateController.instance().createSession();		
		Query wuQuery = hibernateSession.createQuery(hqlQuery);
		List results = wuQuery.list();
		
		associatedWorkUnits = new WorkUnit[results.size()];
		
		if (results.size() > 0) {
			// Convert the reults into a Result array.
			for (int ii = 0; ii < results.size(); ii++)
				associatedWorkUnits[ii] = (WorkUnit)results.get(ii);
		}
		
		try {
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			
		}
		
		return associatedWorkUnits;
	}
	
	/**
	 * Designates a canonical Result by removing all related results from 
	 * the ingress and pending queues, adding the canonical result to 
	 * the valid queue, and retiring the canonical result's work unit.
	 * @param canonicalResult The canonical result to select.
	 */
	public synchronized void selectCanonicalResult(Result canonicalResult) {
		// Find all ingress and pending results for the canonical result's work unit.
		Result[] ingResults = getAssociatedIngressResults(canonicalResult.getWorkUnit());
		Result[] penResults = getAssociatedPendingResults(canonicalResult.getWorkUnit());
		
		// Remove all ingress results except the canonical result.
		for (Result r : ingResults) {
			if (r.equals(canonicalResult) == false) {
				ingressResults.remove(r);	// Remove from the queue.
				r.delete();					// Remove from the database.
			}
		}

		// Remove all pending results except the canonical result.
		for (Result r : penResults) {
			if (r.equals(canonicalResult) == false) {
				pendingResults.remove(r);	// Remove from the queue.
				r.delete();					// Remove from the database.
			}
		}
		
		// Add the canonical result to the valid result queue.		
		moveToValidResultQueue(canonicalResult);
		
		// Remove the result's work unit from the canonical queue.
		canonicalWorkUnits.remove(canonicalResult.getWorkUnit());
		
		// Retire the result's work unit.
		moveToRetiredWorkQueue(canonicalResult.getWorkUnit());
	}
	
	/**
	 * Gets the next work unit to be processed.  First, it will try to find  
	 * a work unit in the pending work queue that has not already been sent 
	 * to the given volunteer.  If there are no satisfiable work units in 
	 * the pending work queue, it will try to return a work unit from the 
	 * ingress work queue.  If there are no work units in the ingress work 
	 * queue, an EmptyQueueException will be thrown.
	 * @param v The volunteer who will receive it.
	 * @return The next work unit to be processed.
	 * @throws EmptyQueueException
	 */
	public synchronized WorkUnit getWorkUnit(Volunteer v) throws EmptyQueueException {
		boolean foundWorkUnit = false;
		WorkUnit wu = null;
		
		/*
		 * First, determine whether this user will get a spot check WorkUnit
		 */
		if (new Random(new Date().getTime()).nextFloat() <= spotCheckProbability) {
			if (spotCheckResults.size() > 0) {
				// Get a random spot check
				int spotCheckIndex = new Random(new Date().getTime()).nextInt(spotCheckResults.size());
				int index = 0;
				for (Result spotCheckResult : spotCheckResults) {
					if (index++ == spotCheckIndex) {
						wu = spotCheckResult.getWorkUnit();
						foundWorkUnit = true;
						break;
					}
				}
			}
		}
		
		/*
		 * Next, look through the pending work queue for work units that
		 * have not yet been sent to the given user.
		 */ 
		Iterator<WorkUnit> workIterator = pendingWorkUnits.iterator();
		while ((workIterator.hasNext() == true) && (foundWorkUnit == false)) {
			boolean assignedToUser = false;
			WorkUnit candidateWU = workIterator.next();
			Iterator<Volunteer> userIterator = candidateWU.getUsers().iterator();
			
			// Compare user 'v' to all users in the WorkUnit
			while (userIterator.hasNext() == true) {
				if (v.equals(userIterator.next())) {
					assignedToUser = true;
					break;
				}
			}

			if (assignedToUser == false) {
				// Found a pending WorkUnit that this user has not yet received.
				wu = candidateWU;
				
				/*
				 * Now, check whether the work unit has already been assigned minNumResults times.
				 * If it has NOT been assigned at least minNumResults times or it has expired, it 
				 * can be assigned to the current user.  Otherwise, optimistically assume that it 
				 * will be retired before it expires and just requeue it. 
				 */
				if (candidateWU.getAssignmentCount() < minNumResults) 
					foundWorkUnit = true;
				
				else if (candidateWU.hasExpired() == true) {
					// Find the total number of returned results for this WU.
					int numReturnedResults = getAssociatedIngressResults(candidateWU).length + getAssociatedPendingResults(candidateWU).length;
					
					// This WU might have only timed out due to validation delays, so only 
					// assign this WU again if there aren't enough results in the queues.
					if (numReturnedResults < minNumResults) {
						// Assign this WU to (minNumResults - numReturnedResults) additional users.
						SysLogger.println("Work unit " + candidateWU.getTaskID() + " has expired.  Reassigning this work unit to " + v.getUserID() + ".");
						foundWorkUnit = true;
						candidateWU.setAssignmentCount(minNumResults - numReturnedResults);
					}
					
					// Reset the expiration date.
					candidateWU.resetExpirationDate();
					
					// Save changes to the work unit.
					candidateWU.update();
				}
				
				// Cycle the WorkUnit to the end of the pending work queue to improve fairness.
				moveToPendingWorkQueue(pendingWorkUnits, wu);
			}
		}

		/*
		 * If no pending WorkUnit was found that this user has not yet received,
		 * try to get a new WorkUnit from the ingress queue.
		 */ 
		if (foundWorkUnit == false) {
			try {
				WorkUnit nextWU = ingressWorkUnits.first();
				wu = nextWU;
				moveToPendingWorkQueue(ingressWorkUnits, wu);
			} catch (NoSuchElementException e) {
				// No valid WorkUnit could be found in any of the queues.
				v.setWaiting(true);		// This user is now waiting for a work unit.
				v.update();
				throw new EmptyQueueException("All work queues are empty.");
			}
		}
		
		// Add the volunteer to the set of users associated with this WorkUnit
		wu.assignUser(v);
		wu.update();
		
		// This user is no longer waiting.
		if (v.getWaiting() == true) {
			v.setWaiting(false);
			v.update();
		}
		
		return wu;
	}
	
	/**
	 * Gets the first work unit that has enough Results.
	 * @return The first work unit in the canonical work queue.
	 * @throws EmptyQueueException
	 */
	public synchronized WorkUnit getCanonicalWorkUnit() throws EmptyQueueException {
		if (canonicalWorkUnits.isEmpty() == true)
			throw new EmptyQueueException("There are no canonical work units.");
		else {
			// Get the first canonical work unit, then cycle it to the end of the queue.
			WorkUnit canonicalWU = canonicalWorkUnits.first();
			canonicalWorkUnits.remove(canonicalWU);
			canonicalWorkUnits.add(canonicalWU);
			return canonicalWU;
		}
	}
	
	/**
	 * Gets the first result that needs to be validated.
	 * @return The first result in the ingress queue.
	 * @throws EmptyQueueException
	 */
	public synchronized Result getResult() throws EmptyQueueException {
		if (ingressResults.isEmpty() == true)
			throw new EmptyQueueException("There are no ingress results.");
		else {
			// Get the first ingress result, then cycle it to the end of the queue.
			Result r = ingressResults.first();
			ingressResults.remove(r);
			ingressResults.add(r);
			return r;
		}
	}
	
	/**
	 * Checks whether sufficient spot checks have been computed.
	 * If not, it will start a new thread to compute them.
	 */
	private void verifySpotChecks() {
		if (spotCheckResults.size() < minNumSpotChecks) {
			// Start the spot check thread.
			SysLogger.println("Transitioner: Starting Spot Check Thread.");
			System.out.println("Transitioner: Starting Spot Check Thread.");
			spotChecker = new SpotCheckThread(
					ingressWorkUnits, 
					spotCheckWorkUnits, 
					spotCheckResults, 
					minNumSpotChecks);
			
			spotChecker.start();
		}
	}
	
	/**
	 * Restores all queues from the database. 
	 */
	private void restoreFromDatabase() {
		// Retrieve all Tasks from the database.
		WorkUnit[] foundIngressWorkUnits = WorkUnit.findAllIngress();
		WorkUnit[] foundPendingWorkUnits = WorkUnit.findAllPending();
		WorkUnit[] foundSpotCheckWorkUnits = WorkUnit.findAllSpotCheck();
		WorkUnit[] foundCanonicalWorkUnits = getWorkUnitsWithEnoughResults();
		Result[] foundIngressResults = Result.findAllIngress();
		Result[] foundPendingResults = Result.findAllPending();
		Result[] foundSpotCheckResults = Result.findAllSpotCheck();
		
		// If any were found, build the queues with the found tasks.
		if (foundIngressWorkUnits.length > 0)
			ingressWorkUnits.addAll(foundIngressWorkUnits);
		if (foundPendingWorkUnits.length > 0)
			pendingWorkUnits.addAll(foundPendingWorkUnits);
		if (foundSpotCheckWorkUnits.length > 0)
			spotCheckWorkUnits.addAll(foundSpotCheckWorkUnits);
		if (foundCanonicalWorkUnits.length > 0)
			canonicalWorkUnits.addAll(foundCanonicalWorkUnits);
		if (foundIngressResults.length > 0)
			ingressResults.addAll(foundIngressResults);
		if (foundPendingResults.length > 0)
			pendingResults.addAll(foundPendingResults);
		if (foundSpotCheckResults.length > 0)
			spotCheckResults.addAll(foundSpotCheckResults);
	}
	
	/**
	 * Performs any necessary tasks to safely shut down the Transitioner.
	 */
	public void shutdown() {
		// Kill the spotchecker thread, if it's running.
		if (spotChecker != null) {
			SysLogger.println("Transitioner: Shutting down Spot Check Thread.");
			System.out.println("Transitioner: Shutting down Spot Check Thread.");
			spotChecker.shutdown();
			SysLogger.println("Transitioner: Spot Check Thread Halted.");
			System.out.println("Transitioner: Spot Check Thread Halted.");
		}
	}
}
