package edu.wpi.prc.server.transitioner;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.server.EmptyQueueException;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.task.queue.MultiPriorityQueue;
import edu.wpi.prc.task.queue.TransitionerQueue;
import edu.wpi.prc.user.Volunteer;

/**
 * The RemoteTransitionerProxy is used when the Transitioner component is executed 
 * on a different machine than the PRCServer.  The RemoteTransitionerProxy uses 
 * XML-RPC calls to communicate with the Transitioner process on another 
 * machine.  If the Transitioner will be run on the same machine, use the 
 * LocalTransitioner instead.
 * @author James Baldassari
 */
public class RemoteTransitionerProxy implements Transitioner {
	private static final int RETRY_DELAY = 10000;		// Wait for 10 seconds before retrying a failed RPC.
	private XmlRpcClient xmlrpcClient = null;
	
	/**
	 * Private constructor ensures that only one instance of RemoteTransitionerProxy exists.
	 * Use the getTransitioner() method to get access to the RemoteTransitionerProxy.
	 * @throws MalformedURLException 
	 */
	public RemoteTransitionerProxy(String transitionerURI) throws MalformedURLException {
		xmlrpcClient = new XmlRpcClient(transitionerURI);
	}

	/**
	 * @return Returns the ingressWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getIngressWorkQueue() {
		return getWorkQueue("IngressWorkQueue");
	}
	
	/**
	 * @return Returns the pendingWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getPendingWorkQueue() {
		return getWorkQueue("PendingWorkQueue");
	}
	
	/**
	 * @return Returns the work units that have enough results to select a canonical result.
	 */
	public TransitionerQueue<WorkUnit> getCanonicalWorkQueue() {
		return getWorkQueue("CanonicalWorkQueue");
	}

	/**
	 * @return Returns the retiredWorkQueue.
	 */
	public TransitionerQueue<WorkUnit> getRetiredWorkQueue() {
		return getWorkQueue("RetiredWorkQueue");
	}

	/**
	 * @return Returns the ingressResultQueue.
	 */
	public TransitionerQueue<Result> getIngressResultQueue() {
		return getResultQueue("IngressResultQueue");
	}
	
	/**
	 * @return Returns the pendingResultQueue.
	 */
	public TransitionerQueue<Result> getPendingResultQueue() {
		return getResultQueue("PendingResultQueue");
	}

	/**
	 * @return Returns the validResultQueue.
	 */
	public TransitionerQueue<Result> getValidResultQueue() {
		return getResultQueue("ValidResultQueue");
	}
	
	/**
	 * @return Returns the retiredResultQueue.
	 */
	public TransitionerQueue<Result> getRetiredResultQueue() {
		return getResultQueue("RetiredResultQueue");
	}

	/**
	 * Adds a work unit to the tail of the ingress work queue.
	 * @param wu The work unit to add.
	 */
	public void addIngressWorkUnit(WorkUnit wu) {
		Vector<String> params = new Vector<String>();
		params.add(wu.getTaskID());
		sendRPC("addIngressWorkUnit", params);
	}

	/**
	 * Adds a result to the tail of the pending result queue.
	 * @param r The result to add.
	 */
	public void addIngressResult(Result r) {
		Vector<String> params = new Vector<String>();
		params.add(r.getTaskID());
		sendRPC("addIngressResult", params);
	}
	
	/**
	 * Finds all ingress Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated ingress Results.
	 * @return Array of ingress Results associated with the given WorkUnit or null if none were found.
	 */
	@SuppressWarnings("unchecked")
	public Result[] getAssociatedIngressResults(WorkUnit wu) {
		return getAssociatedResults(wu, "Ingress");
	}
	
	/**
	 * Finds all pending Results associated with a given WorkUnit.
	 * @param wu The WorkUnit for which to find all associated pending Results.
	 * @return Array of pending Results associated with the given WorkUnit or null if none were found.
	 */
	public Result[] getAssociatedPendingResults(WorkUnit wu) {
		return getAssociatedResults(wu, "Pending");
	}

	/**
	 * Gets the next work unit to be processed.  First, it will try to find  
	 * a work unit in the pending work queue that has not already been sent 
	 * to the given volunteer.  If there are no satisfiable work units in 
	 * the pending work queue, it will try to return a work unit from the 
	 * ingress work queue.  If there are no work units in the ingress work 
	 * queue, an EmptyQueueException will be thrown.
	 * @param v The volunteer who will receive t
	 * @return The next work unit to be processed.
	 * @throws EmptyQueueException
	 */
	public WorkUnit getWorkUnit(Volunteer v) throws EmptyQueueException {
		boolean success = false;
		String taskID = null;
		Vector<String> params = new Vector<String>();
		params.add(v.getUserID());
		params.add(v.getEMailAddress());
		
		while (success == false) {
			Object xmlrpcReturn = sendRPC("getWorkUnit", params);
			try {
				taskID = (String)xmlrpcReturn;
				success = true;
			} catch (ClassCastException e) {
				try {
					Exception transitionerException = (Exception)xmlrpcReturn;
					throw new EmptyQueueException(transitionerException.getMessage());
				} catch (ClassCastException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkUnit: " + e1 + ".");	
				}
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkUnit: " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkUnit: " + e1 + ".");
				}
			}
		}
		
		// Now we have the task ID of the work unit to return, so just query the database for it.
		WorkUnit search = TaskFactory.createWorkUnit(taskID);
		WorkUnit wu = null;
		try {
			wu = search.find()[0];
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkUnit: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkUnit: " + e + ".");
			e.printStackTrace();
		}
		
		return wu;
	}
	
	/**
	 * Gets the first WorkUnit that has enough Results.
	 * @return The first WorkUnit in the canonical work queue.
	 * @throws EmptyQueueException
	 */
	public WorkUnit getCanonicalWorkUnit() throws EmptyQueueException {
		boolean success = false;
		String taskID = null;
		
		while (success == false) {
			Object xmlrpcReturn = sendRPC("getCanonicalWorkUnit", null);
			try {
				taskID = (String)xmlrpcReturn;
				success = true;
			} catch (ClassCastException e) {
				try {
					Exception transitionerException = (Exception)xmlrpcReturn;
					throw new EmptyQueueException(transitionerException.getMessage());
				} catch (ClassCastException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getCanonicalWorkUnit: " + e1 + ".");	
				}
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy.getCanonicalWorkUnit: " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getCanonicalWorkUnit: " + e1 + ".");
				}
			}
		}
		
		// Now we have the task ID of the work unit to return, so just query the database for it.
		WorkUnit search = TaskFactory.createWorkUnit(taskID);
		WorkUnit wu = null;
		try {
			wu = search.find()[0];
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getCanonicalWorkUnit: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getCanonicalWorkUnit: " + e + ".");
			e.printStackTrace();
		}
		
		return wu;
	}
	
	/**
	 * Gets the first result that needs to be validated.
	 * @return The first result in the ingress queue.
	 * @throws EmptyQueueException
	 */
	public synchronized Result getResult() throws EmptyQueueException {
		boolean success = false;
		String taskID = null;
		
		while (success == false) {
			Object xmlrpcReturn = sendRPC("getResult", null);  
			try {
				taskID = (String)xmlrpcReturn;
				success = true;
			} catch (ClassCastException e) {
				try {
					Exception transitionerException = (Exception)xmlrpcReturn;
					throw new EmptyQueueException(transitionerException.getMessage());
				} catch (ClassCastException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getResult: " + e1 + ".");	
				}
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy.getResult: " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getResult: " + e1 + ".");
				}
			}
		}
		
		// Now we have the task ID of the result to return, so just query the database for it.
		Result search = TaskFactory.createResult(taskID);
		Result r = null;
		try {
			r = search.find()[0];
		} catch (ArrayIndexOutOfBoundsException e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getResult: Could not find work unit " + taskID + ".");
		} catch (Exception e) {
			SysLogger.println("ERROR: RemoteTransitionerProxy.getResult: " + e + ".");
			e.printStackTrace();
		}
		
		return r;
	}
	
	/**
	 * Marks a Result as valid.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit, 
	 * the Result is moved from the ingress result queue to 
	 * the pending result queue.  If the minimum number of 
	 * valid results for the Result's associated WorkUnit has 
	 * been reached, all Results for that WorkUnit are removed 
	 * from the pending result queue, a canonical result is chosen, 
	 * and the canonical result is moved to the valid result queue. 
	 * @param validResult The result from the ingress result queue to be marked valid.
	 */
	public synchronized void markResultValid(Result validResult) {
		Vector<String> params = new Vector<String>();
		params.add(validResult.getTaskID());
		sendRPC("markResultValid", params);
	}
	
	/**
	 * Marks a result as invalid.  This will remove the result 
	 * from the ingress queue and delete it from the database.  
	 * If the result is a spot check result, additional steps 
	 * may be taken depending on the project's spot check 
	 * failure action.
	 * @param invalidResult The result to be marked invalid.
	 */
	public synchronized void markResultInvalid(Result invalidResult) {
		Vector<String> params = new Vector<String>();
		params.add(invalidResult.getTaskID());
		sendRPC("markResultInValid", params);
	}
	
	/**
	 * Designates a canonical Result by removing all related results from 
	 * the ingress and pending queues, adding the canonical result to 
	 * the valid queue, and retiring the canonical result's work unit.
	 * @param canonicalResult The canonical result to select.
	 */
	public synchronized void selectCanonicalResult(Result canonicalResult) {
		Vector<String> params = new Vector<String>();
		params.add(canonicalResult.getTaskID());
		sendRPC("selectCanonicalResult", params);
	}
	
	/**
	 * Gets a work queue from the remote transitioner.
	 * @param queueType The type of queue to get: IngressWorkQueue, PendingWorkQueue, CanonicalWorkQueue, RetiredWorkQueue
	 * @return Returns the work queue that was returned by the remote transitioner.
	 */
	@SuppressWarnings("unchecked")
	private TransitionerQueue<WorkUnit> getWorkQueue(String queueType) {
		boolean success = false;
		Vector<String> workQueueVector = null;
		TransitionerQueue<WorkUnit> workQueue = new MultiPriorityQueue<WorkUnit>();
		String methodName = "get" + queueType;
		
		// Execute the XML-RPC.
		while (success == false) {
			try {
				workQueueVector = (Vector)sendRPC(methodName, null);
				success = true;
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkQueue: " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getWorkQueue: " + e1 + ".");
				}
			}
		}
		
		// Now we have the Vector of work unit task IDs.
		// Iterate over the Vector, adding the corresponding WorkUnits to the set.
		for (String taskID : workQueueVector) {
			WorkUnit search = TaskFactory.createWorkUnit(taskID);
			try {
				WorkUnit wu = search.find()[0];
				workQueue.add(wu);
			} catch (ArrayIndexOutOfBoundsException e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": Could not find work unit " + taskID + ".");
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				e.printStackTrace();
			}
		}
		
		return workQueue;
	}
	
	/**
	 * Gets a result queue from the remote transitioner.
	 * @param queueType The type of queue to get: IngressResultQueue, PendingResultQueue, ValidResultQueue, or RetiredResultQueue
	 * @return Returns the result queue that was returned by the remote transitioner.
	 */
	@SuppressWarnings("unchecked")
	private TransitionerQueue<Result> getResultQueue(String queueType) {
		boolean success = false;
		Vector<String> resultQueueVector = null;
		TransitionerQueue<Result> resultQueue = new MultiPriorityQueue<Result>();
		String methodName = "get" + queueType;
		
		while (success == false) {
			try {
				resultQueueVector = (Vector)sendRPC(methodName, null);
				success = true;				
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy.getResultQueue: " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy.getResultQueue: " + e1 + ".");
				}
			}
		}
		
		// Now we have the Vector of result task IDs.
		// Iterate over the Vector, adding the corresponding Results to the set.
		for (String taskID : resultQueueVector) {
			Result search = TaskFactory.createResult(taskID);
			try {
				Result r = search.find()[0];
				resultQueue.add(r);
			} catch (ArrayIndexOutOfBoundsException e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": Could not find result " + taskID + ".");
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				e.printStackTrace();
			}
		}
		
		return resultQueue;
	}
	
	/**
	 * Gets the set of results associated with the given work unit.
	 * @param wu The work unit for which to find associated results.
	 * @param resultType The type of results to get: Ingress or Pending.
	 * @return Returns the set of results returned by the remote transitioner.
	 */
	@SuppressWarnings("unchecked")
	private Result[] getAssociatedResults(WorkUnit wu, String resultType) {
		boolean success = false;
		Vector<String> resultVector = null;
		Vector<String> params = new Vector<String>();
		params.add(wu.getTaskID());
		String methodName = "getAssociated" + resultType + "Results";
		
		while (success == false) {
			try {
				resultVector = (Vector)sendRPC(methodName, params);
				success = true;				
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e1 + ".");
				}
			}
		}
		
		// Now we have the Vector of result task IDs.
		// Iterate over the Vector, adding the corresponding Results to the set.
		Result[] associatedResults = new Result[resultVector.size()];
		for (int ii = 0; ii < resultVector.size(); ii++) {
			String taskID = resultVector.get(ii);
			Result search = TaskFactory.createResult(taskID);
			try {
				Result r = search.find()[0];
				associatedResults[ii] = r;
			} catch (ArrayIndexOutOfBoundsException e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": Could not find result " + taskID + ".");
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				e.printStackTrace();
			}
		}
		
		return associatedResults;
	}
	
	/**
	 * Executes an XML-RPC on the remote transitioner.
	 * @param methodName The name of the method to execute.
	 * @param params Parameters to pass to the method.
	 * @return Returns whatever value was returned by the XML-RPC.
	 */
	private Object sendRPC(String methodName, Vector params) {
		boolean success = false;
		Object xmlrpcReturn = null;
		Vector rpcParameters = (params != null) ? params : new Vector();
		while (success == false) {
			try {
				xmlrpcReturn = xmlrpcClient.execute("transitioner." + methodName, rpcParameters);
				success = true;
			} catch (XmlRpcException e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e1 + ".");
				}
			} catch (IOException e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e1 + ".");
				}
			} catch (Exception e) {
				SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e + ".");
				SysLogger.println("Waiting for " + (RETRY_DELAY / 1000) + " seconds before retry attempt.");
				try {
					Thread.sleep(RETRY_DELAY);
				} catch (InterruptedException e1) {
					SysLogger.println("ERROR: RemoteTransitionerProxy." + methodName + ": " + e1 + ".");
				}
			}
		}
		
		return xmlrpcReturn;
	}
	
	/**
	 * Performs any necessary tasks to safely shut down the Transitioner.
	 */
	public void shutdown() {
		sendRPC("shutdown", null);
	}
}
