package edu.wpi.prc.task;

import java.util.Date;
import java.util.LinkedList;
import java.util.List;

/**
 * The Task class stores information about each task (WorkUnit or Result), 
 * including the task's ID, the users who are responsible for this task, and the 
 * data the task contains.
 * Note: this class has a natural ordering that is inconsistent with equals.
 * @author James Baldassari
 */
public abstract class Task implements Prioritizable {
	public static final int MIN_PRIORITY = 0;
	private static final int PARTITION_LENGTH = 65536;	// 64K
	private long id;								// Hibernate ID
	private String taskID = null;
	private List<byte[]> dataPartitions = new LinkedList<byte[]>();
	private String stateStr = null;
	private Date creationDate = null;
	private int priority = MIN_PRIORITY; 		// Higher values -> higher priorities.
	
	/**
	 * @return Returns the id.
	 */
	protected long getId() {
		return id;
	}

	/**
	 * @param id The id to set.
	 */
	protected void setId(long id) {
		this.id = id;
	}
	
	/**
	 * @return Returns the list of data partitions (required for Hibernate to work).
	 */
	protected List<byte[]> getDataPartitions() {
		return dataPartitions;
	}
	
	/**
	 * Sets the list of data partitions (required for Hibernate to work).
	 */
	protected void setDataPartitions(List<byte[]> dataPartitions) {
		this.dataPartitions = dataPartitions;
	}

	/**
	 * @return Returns the data.
	 */
	public byte[] getData() {
		/*
		 * This method actually reassembes all of the data partitions
		 * into a single byte array.  The reason it is necessary to 
		 * partition the data into smaller byte arrays is that databases
		 * like MySQL and PostgreSQL can't seem to handle a single binary
		 * field of length 4096.  So we split up the data into 64K chunks. 
		 */
		
		int numPartitions = dataPartitions.size();
		
		// There is no data, so return an empty byte array.
		if (numPartitions == 0)
			return new byte[0];
		
		else {
			// All partitions except the last will be of length PARTITION_LENGTH.
			int dataLength = (numPartitions - 1) * PARTITION_LENGTH;
			
			// The last partition may have length between 1 and PARTITION_LENGTH.
			dataLength += dataPartitions.get(numPartitions - 1).length;
			
			// Allocate a new byte array big enough to hold all partitions.
			byte[] data = new byte[dataLength];
			
			// Copy over the data from each partition to the byte array.
			int dataIndex = 0;
			for (byte[] partition : dataPartitions) {
				for (int ii = 0; ii < partition.length; ii++) {
					data[dataIndex] = partition[ii];
					dataIndex++;
				}
			}
			
			// Return the byte array containing all partitions.
			return data;
		}
	}
	
	/**
	 * @param data The data to set.
	 */
	public void setData(byte[] data) {
		/*
		 * This method actually partitions the data into several smaller 
		 * single byte arrays.  The reason it is necessary to 
		 * partition the data into smaller byte arrays is that databases
		 * like MySQL and PostgreSQL can't seem to handle a single binary
		 * field of length 4096.  So we split up the data into 64K chunks. 
		 */
		
		if (data == null) {
			dataPartitions = new LinkedList<byte[]>();
		}
		
		else if (data.length == 0) {
			dataPartitions = new LinkedList<byte[]>();
		}
		
		else {
			// Determine how many partitions of lengh PARTITION_LENGTH are needed to store the data.
			int numFullPartitions = data.length / PARTITION_LENGTH;
			
			// Copy the data into the partitions.
			int dataIndex = 0;
			for (int ii = 0; ii < numFullPartitions; ii++) {
				// Allocate the new partition.
				byte[] partition = new byte[PARTITION_LENGTH];
				
				// Copy the data to the new partition.
				for (int jj = 0; jj < partition.length; jj++) {
					partition[jj] = data[dataIndex];
					dataIndex++;
				}
				
				// Add this partition to the list of partitions.
				dataPartitions.add(partition);
			}
			
			// There may be a last partition with length between 1 and PARTITION_LENGTH-1
			int lastPartitionLength = data.length % PARTITION_LENGTH;
			if (lastPartitionLength > 0) {
				byte[] partition = new byte[lastPartitionLength];
				for (int ii = 0; ii < partition.length; ii++) {
					partition[ii] = data[dataIndex];
					dataIndex++;
				}
				dataPartitions.add(partition);
			}
		}
	}
	
	/**
	 * @return Returns the taskID.
	 */
	public String getTaskID() {
		return taskID;
	}
	
	/**
	 * @param taskID The taskID to set.
	 */
	public void setTaskID(String taskID) {
		this.taskID = taskID;
	}

	/**
	 * @return Returns the stateStr.
	 */
	protected String getStateStr() {
		return stateStr;
	}

	/**
	 * @param taskStateStr The stateStr to set.
	 */
	protected void setStateStr(String taskStateStr) {
		this.stateStr = taskStateStr;
	}
	
	/**
	 * @return Returns the current state of this Task.
	 */
	public TaskState getState() {
		return Enum.valueOf(TaskState.class, stateStr);
	}
	
	/**
	 * Sets the current state of this Task.
	 * @param state The state to set.
	 */
	public void setState(TaskState state) {
		this.stateStr = state.toString();
	}

	/**
	 * @return Returns the creationDate.
	 */
	public Date getCreationDate() {
		return creationDate;
	}

	/**
	 * @param creationDate The creationDate to set.
	 */
	protected void setCreationDate(Date creationDate) {
		this.creationDate = creationDate;
	}

	/**
	 * @return Returns the priority.
	 */
	public int getPriority() {
		return priority;
	}

	/**
	 * @param priority The priority to set.
	 */
	public void setPriority(int priority) {
		if (priority < MIN_PRIORITY)
			priority = MIN_PRIORITY;
		else
			this.priority = priority;
	}

	/**
	 * Inserts this Task into the database as a new record.
	 */
	public abstract void save();
	
	/**
	 * Updates this Task in the database.
	 */
	public abstract void update();
	
	/**
	 * Deletes this Task from the database.
	 */
	public abstract void delete();

	/**
	 * Tests whether this Task is equal to another.
	 * @param o The object to compare this Task to.
	 */
	public boolean isEqual(Object o) {
		try {
			Task t = (Task)o;
			return this.taskID.equals(t.getTaskID());
		} catch (Exception e) {
			return false;
		}
	}
}
