package edu.wpi.prc.task;

import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * Creates instances of the WorkUnit and Result classes.
 * Automatically assigns a unique task ID to each instance.
 * Note: created instances are not automatically stored to the database.
 * @author James Baldassari
 */
public class TaskFactory {	
	private static final String TASK_ID_FORMAT = new String("yyyy-MM-dd-HH-mm-ss-SSS");
	private static final String WORK_UNIT_PREFIX = "WU_";
	private static final String RESULT_PREFIX = "Re_";
	private static final int MAX_END_ID = 99;
	private static int endID = 0;
	private static SimpleDateFormat taskFormat = new SimpleDateFormat(TASK_ID_FORMAT);
	
	/**
	 * @return Returns a new WorkUnit with a unique task ID.
	 */
	public static WorkUnit createWorkUnit() {
		WorkUnit wu = new WorkUnit();
		wu.setTaskID(generateWorkUnitID());
		wu.setCreationDate(new Date());
		return wu;
	}
	
	/**
	 * Creates a new WorkUnit with the given task ID.
	 * @param taskID The task ID to set for the new WorkUnit.
	 * @return Returns a new WorkUnit with a unique taskID.
	 */
	public static WorkUnit createWorkUnit(String taskID) {
		return new WorkUnit(taskID);
	}
	
	/**
	 * Creates a new WorkUnit.  The WorkUnit's state is set to ingress.
	 * @return Returns a new ingress WorkUnit with a unique task ID.
	 */
	public static WorkUnit createIngressWorkUnit() {
		WorkUnit wu = createWorkUnit();
		wu.setState(TaskState.Ingress);
		return wu;
	}

	/**
	 * Creates a new WorkUnit with the given task ID.  The WorkUnit's state is set to ingress.
	 * @param taskID The task ID to set for the new WorkUnit.
	 * @return Returns a new ingress WorkUnit with the given task ID.
	 */
	public static WorkUnit createIngressWorkUnit(String taskID) {
		WorkUnit wu = createIngressWorkUnit();
		wu.setTaskID(taskID);
		return wu;
	}
	
	/**
	 * @return Returns a new Result with a unique task ID.
	 */
	public static Result createResult() {
		Result r = new Result();
		r.setTaskID(generateResultID());
		r.setCreationDate(new Date());
		return r;
	}
	
	/**
	 * Creates a new Result with the given task ID.
	 * @param taskID The task ID to set for the new Result.
	 * @return Returns a new Result with the given task ID.
	 */
	public static Result createResult(String taskID) {
		return new Result(taskID);
	}
	
	/**
	 * Creates a new Result.  The Result's state is set to ingress.
	 * @return Returns a new ingress Result with a unique task ID.
	 */
	public static Result createIngressResult() {
		Result r = createResult();
		r.setState(TaskState.Ingress);
		return r;
	}
	
	/**
	 * Creates a new Result with the given task ID.  The Result's state is set to ingress.
	 * @param taskID The task ID to set for the new Result.
	 * @return Returns a new ingress Result with the given task ID.
	 */
	public static Result createIngressResult(String taskID) {
		Result r = createIngressResult();
		r.setTaskID(taskID);
		return r;
	}
	
	/**
	 * @return Returns a unique task ID based on the current time and an integer [0,99].
	 */
	private static String generateTaskID(String prefix) {
		String taskID = null;
		boolean isUnique = false;
		while (isUnique == false) {
			taskID = prefix + taskFormat.format(new Date()) + "@" + endID;
			endID++;
			if (endID > MAX_END_ID)
				endID = 0;
			// If there is no task in the database with this task ID, we have a unique ID.
			if ((new WorkUnit(taskID).find().length == 0) && (new Result(taskID).find().length == 0))
				isUnique = true;
		}
		return taskID;
	}
	
	/**
	 * @return a unique work unit ID.
	 */
	private static String generateWorkUnitID() {
		return generateTaskID(WORK_UNIT_PREFIX);
	}
	
	/**
	 * @return a unique result ID.
	 */
	private static String generateResultID() {
		return generateTaskID(RESULT_PREFIX);
	}
}
