package edu.wpi.prc.task;

import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Expression;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.persistence.SysLogger;
import edu.wpi.prc.user.Volunteer;

/**
 * A WorkUnit is a type of Task that contains raw data to 
 * be analyzed by the science application.  It can be sent 
 * to one or more clients.
 * @author James Baldassari
 */
public class WorkUnit extends Task {
	public static final long DEFAULT_EXPIRATION_TIME = 86400000;		// Default expiration time is 24 hours (in milliseconds).
	private int points = 1;
	private int numValidResults = 0;
	private Set<Volunteer> users = new HashSet<Volunteer>();
	private int assignmentCount = 0;		// Number of times this WU has been assigned, which may be < users.size() if this WU expired. 
	private Date expirationDate = null;	// The date by which this work unit should be retired.
	private long expirationDelta = 0;	// Number of seconds this work unit should spend in the ingress/pending queues.
	
	/**
	 * Default constructor is protected so that the WorkUnit will either be created
	 * by the TaskFactory or the public constructor that takes a task ID will be 
	 * called.  This restriction ensures that the WorkUnit will always have a valid 
	 * task ID.
	 */
	protected WorkUnit() {
		
	}
	
	/**
	 * Constructor that takes a task ID.
	 * @param taskID The task ID to set.
	 */
	protected WorkUnit(String taskID) {
		assert(taskID != null);
		super.setTaskID(taskID);
	}

	/**
	 * @return Returns the points that will be added to the user's score
	 * when a Result is returned for this WorkUnit.
	 */
	public int getPoints() {
		return points;
	}

	/**
	 * @param points The points to set.
	 */
	public void setPoints(int points) {
		if (points < 0)
			this.points = 0;
		else
			this.points = points;
	}

	/**
	 * @return Returns the numValidResults.
	 */
	public int getNumValidResults() {
		return numValidResults;
	}

	/**
	 * @param numValidResults The numValidResults to set.
	 */
	protected void setNumValidResults(int numValidResults) {
		this.numValidResults = numValidResults;
	}
	
	/**
	 * Increments the number of valid results for this work unit.
	 */
	public void incrementNumValidResults() {
		numValidResults++;
	}
	
	/**
	 * @return Returns the users.
	 */
	public Set<Volunteer> getUsers() {
		return users;
	}
	
	/**
	 * @param users The users to set.
	 */
	protected void setUsers(Set<Volunteer> users) {
		this.users = users;
	}
	
	/**
	 * Add a user to the set of users associated with this work unit, and 
	 * increments the assignment count.
	 * @param user The user to add.
	 */
	public void assignUser(Volunteer user) {
		users.add(user);
		assignmentCount++;
	}
	
	/**
	 * Removes all users from this work unit. 
	 */
	public void removeAllusers() {
		users.clear();
	}
	
	/**
	 * Tests whether this work unit has been assigned to the given user.
	 * @param user The user to search for.
	 * @return True if the given user is associated with this work unit, false otherwise.
	 */
	public boolean hasUser(Volunteer user) {
		for (Volunteer v : users) {
			if (v.equals(user))
				return true;
		}
		return false;
	}
	
	/**
	 * Remove a user from the set of users associated with this work unit.
	 * @param user The user to remove.
	 * @return The user that was removed, or null if the user did not appear in the set.
	 */
	public Volunteer removeUser(Volunteer user) {
		Volunteer removedUser = null;
		if (users.contains(user)) {
			Iterator<Volunteer> vIter = users.iterator();
			while (vIter.hasNext()) {
				Volunteer v = vIter.next();
				if (v.equals(user)) {
					removedUser = v;
					break;
				}
			}
			users.remove(user);
		}
		return removedUser;
	}
	
	/**
	 * @return Returns the assignmentCount.
	 */
	public int getAssignmentCount() {
		return assignmentCount;
	}
	
	/**
	 * @param assignmentCount The assignmentCount to set.
	 */
	public void setAssignmentCount(int assignmentCount) {
		this.assignmentCount = assignmentCount;
	}
	
	/**
	 * @return Returns the expirationDate.
	 */
	protected Date getExpirationDate() {
		return expirationDate;
	}

	/**
	 * @param expirationDate The expirationDate to set.
	 */
	protected void setExpirationDate(Date expirationDate) {
		this.expirationDate = expirationDate;
	}
	
	/**
	 * @return Returns true if the expiration date for this work unit has passed, false otherwise.
	 */
	public boolean hasExpired() {
		if (expirationDate == null)
			return false;
		else
			return expirationDate.before(new Date());
	}

	/**
	 * @return Returns the expirationDelta.
	 */
	protected long getExpirationDelta() {
		return expirationDelta;
	}
	
	/**
	 * @param expirationDelta The expirationDelta to set.
	 */
	protected void setExpirationDelta(long expirationDelta) {
		this.expirationDelta = expirationDelta;
	}

	/**
	 * Sets the expiration time and, then calculates and sets the expiration date.
	 * NOTE: This method will only set the expiration time if it is greater than zero. 
	 * @param expirationTime The expiration time in seconds to set.
	 */
	public void setExpirationTime(long expirationTime) {
		if (expirationTime > 0) {
			setExpirationDelta(expirationTime);
			Date now = new Date();
			expirationDate = new Date(now.getTime() + (expirationTime * 1000));
		}
		else {
			SysLogger.println("WARNING: WorkUnit.setExpirationTime called with expiration time less than zero.");
			SysLogger.println("WARNING: Using WorkUnit.DEFAULT_EXPIRATION_TIME (" + DEFAULT_EXPIRATION_TIME + " seconds).");
			setExpirationDelta(DEFAULT_EXPIRATION_TIME);
		}
	}

	/**
	 * Sets this work unit's expiration date to the current time plus the expiration time. 
	 * NOTE: This method will only reset the expiration date if this work unit has expired. 
	 */
	public void resetExpirationDate() {
		// Only reset the expiration date if this work unit has expired.
		if (hasExpired()) {
			// Get the current time.
			Date now = new Date();
			// Add the expiration time to the current time to determine the next expiration date.
			long adjustedTime = now.getTime() + (expirationDelta * 1000);
			// Set the new expiration date.
			setExpirationDate(new Date(adjustedTime));
		}
	}

	/**
	 * Inserts this WorkUnit into the database as a new record.
	 */
	public void save() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.save(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Updates this WorkUnit in the database.
	 */
	public void update() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.update(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Finds all WorkUnits in the database that match the attributes set in this WorkUnit.
	 * @return An array of all matching WorkUnits.
	 */
	public WorkUnit[] find()  {
		WorkUnit[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			String tID = getTaskID();
			String stateStr = getStateStr();
			Criteria query = hibernateSession.createCriteria(WorkUnit.class);
			
			// Dynamically build the query.
			if (tID != null)
				query.add(Expression.eq("taskID", tID));
			if (stateStr != null)
				query.add(Expression.eq("stateStr", stateStr));
			
			// Execute the query.
			List results = query.list();
			
			// Build array of WorkUnits
			matches = new WorkUnit[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (WorkUnit)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
			e.printStackTrace();
		} finally {
			try {
				if (hibernateSession != null)
					hibernateSession.close();
			} catch (Exception e) {
				
			}
		}
		
		return matches;
	}
	
	/**
	 * Finds all WorkUnits in the database.
	 * @return An array of all WorkUnits in the database.
	 */
	public static WorkUnit[] findAll() {
		WorkUnit[] matches = null;
		Session hibernateSession = HibernateController.instance().createSession();
		
		try {
			Criteria query = hibernateSession.createCriteria(WorkUnit.class);
			
			// Execute the query.
			List results = query.list();
			
			// Build array of WorkUnits
			matches = new WorkUnit[results.size()];
			Iterator iter = results.iterator();
			int ii = 0;
			while (iter.hasNext())
				matches[ii++] = (WorkUnit)iter.next();
			
			// Clean up session.
			hibernateSession.close();
		} catch (HibernateException e) {
			e.printStackTrace();
		} finally {
			try {
				if (hibernateSession != null)
					hibernateSession.close();
			} catch (Exception e) {
				
			}
		}
		
		return matches;
	}
	
	/**
	 * @return Returns an array of all WorkUnits in the Ingress state.
	 */
	public static WorkUnit[] findAllIngress() {
		WorkUnit wu = new WorkUnit();
		wu.setState(TaskState.Ingress);
		return wu.find();
	}
	
	/**
	 * @return Returns an array of all WorkUnits in the Pending state.
	 */
	public static WorkUnit[] findAllPending() {
		WorkUnit wu = new WorkUnit();
		wu.setState(TaskState.Pending);
		return wu.find();
	}
	
	/**
	 * @return Returns an array of all WorkUnits in the Egress state.
	 */
	public static WorkUnit[] findAllEgress() {
		WorkUnit wu = new WorkUnit();
		wu.setState(TaskState.Egress);
		return wu.find();
	}
	
	/**
	 * @return Returns an array of all WorkUnits in the SpotCheck state.
	 */
	public static WorkUnit[] findAllSpotCheck() {
		WorkUnit wu = new WorkUnit();
		wu.setState(TaskState.SpotCheck);
		return wu.find();
	}
	
	/**
	 * Deletes a WorkUnit from the database.
	 */
	public void delete() {
		Session hibernateSession = HibernateController.instance().createSession();
		Transaction tx = null;
		
		try {
			tx = hibernateSession.beginTransaction();
			hibernateSession.delete(this);
			tx.commit();
			
			// Clean up the session.
			hibernateSession.close();
		} catch (HibernateException e) {
			//e.printStackTrace();
			try {
				tx.rollback();
			} catch (Exception e2) {
				// Silent failure.
			} finally {
				try {
					if (hibernateSession != null)
						hibernateSession.close();
				} catch (Exception e3) {
					// Silent failure.
				}
			}
		}
	}
	
	/**
	 * Tests whether this WorkUnit is equal to another.
	 * @param o The object to compare this WorkUnit to.
	 */
	public boolean equals(Object o) {
		try {
			WorkUnit wu = (WorkUnit)o;
			return wu.getTaskID().equals(this.getTaskID());
		} catch (Exception e) {
			return false;
		}
	}
}
