package edu.wpi.prc.task.test;

import java.util.Arrays;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.WorkUnit;
import junit.framework.TestCase;

/**
 * Tests data partitioning in the Task class.
 * @author James Baldassari
 */
public class DataTest extends TestCase {
	private static final int PARTITION_LENGTH = 65536;	// 64K.  NOTE: This will have to change if changed in Task!
	private WorkUnit wu = null;

	/**
	 * The constructor initializes Hibernate.
	 * @param name The name of the test.
	 */
	public DataTest(String name) {
		super(name);
		
		HibernateController.instance().initializeForUnitTests();
	}

	/**
	 * Before every test we start with a clean database by exporting the schema.
	 */
	protected void setUp() throws Exception {
		super.setUp();
		HibernateController.instance().exportSchema();
		wu = TaskFactory.createWorkUnit();
	}
	
	/**
	 * Shuts down Hibernate, then initializes it.
	 */
	private void reinitializeHibernate() {
		HibernateController.instance().shutdown();
		HibernateController.instance().initializeForUnitTests();
	}
	
	/**
	 * Tests getting the data before it has been set and setting a null value for the data.
	 */
	public void testNullData() {
		assertTrue(wu.getData().length == 0);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		wu.setData(null);
		assertTrue(wu.getData().length == 0);
	}
	
	/**
	 * Tests setting the data using a byte array of zero length.
	 */
	public void testZeroData() {
		wu.setData(new byte[0]);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		assert(wu.getData().length == 0);
	}
	
	/**
	 * Tests saving data with length = 1.
	 */
	public void testDataLengthOne() {
		byte[] data = new byte[1];
		data[0] = (byte)2;
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == 1);
		assertTrue(savedData[0] == (byte)2);
	}
	
	/**
	 * Tests saving data with 1 < length < PARTITION_LENGTH.
	 */
	public void testSmallDataLength() {
		int dataLen = PARTITION_LENGTH / 3;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)7);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)7);
		}
	}
	
	/**
	 * Tests saving data with length = PARTITION_LENGTH - 1.
	 */
	public void testOneAlmostCompletePartition() {
		int dataLen = PARTITION_LENGTH - 1;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)6);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)6);
		}
	}
	
	/**
	 * Tests saving data with length = PARTITION_LENGTH.
	 */
	public void testOneCompletePartition() {
		int dataLen = PARTITION_LENGTH;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)5);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)5);
		}
	}
	
	/**
	 * Tests saving data with PARTITION_LENGTH < length < 2 * PARTITION_LENGTH.
	 */
	public void testOneCompleteOneIncompletePartition() {
		int dataLen = PARTITION_LENGTH + 1;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)4);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)4);
		}
	}
	
	/**
	 * Tests saving a 308KB byte array.
	 */
	public void testBigData() {
		int dataLen = 315392;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)3);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)3);
		}
	}
	
	/**
	 * Tests saving a 4MB byte array.
	 */
	public void testReallyBigData() {
		int dataLen = 4194304;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, (byte)1);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		for (byte datum : savedData) {
			assertTrue(datum == (byte)1);
		}
	}
	
	/**
	 * Verfies that the order of the data does not change after it has been saved/retrieved.
	 */
	public void testOrder() {
		int dataLen = 2 * PARTITION_LENGTH;
		byte[] data = new byte[dataLen];
		Arrays.fill(data, 0, PARTITION_LENGTH, (byte)1);
		Arrays.fill(data, PARTITION_LENGTH, 2 * PARTITION_LENGTH, (byte)2);
		wu.setData(data);
		wu.save();
		
		reinitializeHibernate();
		wu = WorkUnit.findAll()[0];
		
		byte[] savedData = wu.getData();
		assertTrue(savedData.length == dataLen);
		
		for (int ii = 0; ii < PARTITION_LENGTH; ii++)
			assertTrue(savedData[ii] == (byte)1);

		for (int ii = PARTITION_LENGTH; ii < 2 * PARTITION_LENGTH; ii++)
			assertTrue(savedData[ii] == (byte)2);
	}
}
