package edu.wpi.prc.task.test;

import java.util.Date;

import edu.wpi.prc.persistence.HibernateController;
import edu.wpi.prc.task.Result;
import edu.wpi.prc.task.TaskFactory;
import edu.wpi.prc.task.TaskState;
import edu.wpi.prc.task.WorkUnit;
import edu.wpi.prc.user.Volunteer;
import junit.framework.TestCase;

/**
 * Tests Result persistence.
 * @author James Baldassari
 */
public class ResultTest extends TestCase {
	private Result r = null;
	private WorkUnit wu = null;
	private Volunteer v = null;

	public ResultTest(String name) {
		super(name);
		
		HibernateController.instance().initializeForUnitTests();
		r = TaskFactory.createResult();
		wu = TaskFactory.createWorkUnit();
		v = new Volunteer();
		
		v.setUserID("jdb");
		v.setEMailAddress("jdb@cs.wpi.edu");
		v.setScore(1);
		v.setJoinDate(new Date());
		v.setLastResultDate(new Date());
		
		byte[] workData = {0, 1, 2, 3};
		wu.setTaskID("T1000");
		wu.assignUser(v);
		wu.setData(workData);
		wu.incrementNumValidResults();
		
		byte[] resultData = {3, 2, 1, 0};
		r.setTaskID("R3000z");
		r.setState(TaskState.Ingress);
		r.setData(resultData);
		r.setVolunteer(v);
		r.setWorkUnit(wu);
	}

	protected void setUp() throws Exception {
		super.setUp();
		HibernateController.instance().exportSchema();
		v.save();
		wu.save();
	}

	/**
	 * Tests saving a WorkUnit to the database.
	 */
	public void testSave() {
		// Save the Result to the database.
		r.save();
		
		// Find the Result in the database.
		Result r2 = TaskFactory.createResult("R3000z");
		r2.setState(TaskState.Ingress);
		Result[] matches = r2.find();
		
		// Verify results.
		assertTrue(matches.length == 1);
		assertTrue(matches[0].getTaskID().equals("R3000z"));
		assertTrue(matches[0].getState() == TaskState.Ingress);
		assertTrue(matches[0].getWorkUnit().getTaskID().equals("T1000"));
		assertTrue(matches[0].getWorkUnit().getNumValidResults() == 1);
		for (int ii = 0; ii < 4; ii++)
			assertTrue(matches[0].getData()[ii] == (byte)(3 - ii));
	}

	/**
	 * Tests updating a Result in the database.
	 */
	public void testUpdate() {
		// Save the Result to the database.
		r.save();
		
		// Find all Results.
		Result[] matches = Result.findAll();
		assertTrue(matches.length == 1);
		
		// Modify the Result and its WorkUnit.
		matches[0].setTaskID("R4000z");
		matches[0].setState(TaskState.Pending);
		matches[0].getWorkUnit().setTaskID("T2000");
		matches[0].getWorkUnit().setState(TaskState.Egress);
		matches[0].getWorkUnit().incrementNumValidResults();
		matches[0].getWorkUnit().update();
		
		// Save the modified Result and cascade changes to its WorkUnit.
		matches[0].update();
		
		// Find all WorkUnits.
		matches = null;
		matches = Result.findAll();
		
		// Verify results.
		assertTrue(matches.length == 1);
		assertTrue(matches[0].getTaskID().equals("R4000z"));
		assertTrue(matches[0].getState() == TaskState.Pending);
		assertTrue(matches[0].getWorkUnit().getTaskID().equals("T2000"));
		assertTrue(matches[0].getWorkUnit().getState() == TaskState.Egress);
		assertTrue(matches[0].getWorkUnit().getNumValidResults() == 2);
	}
	
	/**
	 * Tests deleting a Result from the database.
	 */
	public void testDelete() {
		// Save the Result to the database.
		r.save();
		
		// Find all Results.
		Result[] matches = Result.findAll();
		assertTrue(matches.length == 1);
		
		// Delete the Result.
		r.delete();
		
		// Find all Results.
		matches = null;
		matches = Result.findAll();
		
		// Verify results.
		assertTrue(matches.length == 0);
		
		// Find all WorkUnits and Volunteers.
		WorkUnit[] units = WorkUnit.findAll();
		Volunteer[] vols = Volunteer.findAll();
		
		// Deleting the Result should not have deleted the WorkUnit or Volunteer
		assertTrue(units.length == 1);
		assertTrue(vols.length == 1);
	}
	
}
