package edu.wpi.prc.tools.builder.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;

import com.jeta.forms.components.panel.FormPanel;

import edu.wpi.prc.project.Project;
import edu.wpi.prc.tools.builder.ProjectBuilder;

/**
 * Allows the user to configure the project server address, and the port numbers for the server, client, and science application.
 * @author James Baldassari
 */
public class NetworkConfigurationForm extends GUIForm {
	private static final long serialVersionUID = 1L;
	private GUIForm previousForm = null;
	private FormPanel currentForm = null;
	private GUIForm nextForm = null;
	private Project project = null;
	private SpinnerNumberModel serverPortSpinnerModel = new SpinnerNumberModel();
	private SpinnerNumberModel clientPortSpinnerModel = new SpinnerNumberModel();
	private SpinnerNumberModel sciappPortSpinnerModel = new SpinnerNumberModel();
	private JSpinner serverPortSpinner = null;
	private JSpinner clientPortSpinner = null;
	private JSpinner sciappPortSpinner = null;

	/**
	 * Constructor that takes a reference to the this form's parent form.
	 */
	public NetworkConfigurationForm(GUIForm previousForm) {		
		this.previousForm = previousForm;
		currentForm = new FormPanel("edu/wpi/prc/tools/builder/gui/forms/NetworkConfigurationForm.jfrm");
		setLayout(new BorderLayout());
		add(currentForm, BorderLayout.CENTER);
		this.setTitle("PRC Project Builder: Step 4 of 7");
		
		// Initialize spinners
		serverPortSpinnerModel.setMinimum(1);
		serverPortSpinnerModel.setStepSize(1);
		serverPortSpinnerModel.setMaximum(65536);
		clientPortSpinnerModel.setMinimum(1);
		clientPortSpinnerModel.setStepSize(1);
		clientPortSpinnerModel.setMaximum(65536);
		sciappPortSpinnerModel.setMinimum(1);
		sciappPortSpinnerModel.setStepSize(1);
		sciappPortSpinnerModel.setMaximum(65536);
		
		serverPortSpinner = currentForm.getSpinner(NetworkConfigurationComponents.ID_SERVER_PORT_SPINNER);
		clientPortSpinner = currentForm.getSpinner(NetworkConfigurationComponents.ID_CLIENT_PORT_SPINNER);
		sciappPortSpinner = currentForm.getSpinner(NetworkConfigurationComponents.ID_SCIAPP_PORT_SPINNER);
		
		serverPortSpinner.setModel(serverPortSpinnerModel);
		serverPortSpinner.setEditor(new JSpinner.NumberEditor(serverPortSpinner, "#####"));
		clientPortSpinner.setModel(clientPortSpinnerModel);
		clientPortSpinner.setEditor(new JSpinner.NumberEditor(clientPortSpinner, "#####"));
		sciappPortSpinner.setModel(sciappPortSpinnerModel);
		sciappPortSpinner.setEditor(new JSpinner.NumberEditor(sciappPortSpinner, "#####"));
		
		((JSpinner.DefaultEditor)serverPortSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		((JSpinner.DefaultEditor)clientPortSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		((JSpinner.DefaultEditor)sciappPortSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);

		// Add action listeners.
		currentForm.getButton(NetworkConfigurationComponents.ID_PREVIOUS_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionPreviousClick();
			}
		});
		
		currentForm.getButton(NetworkConfigurationComponents.ID_NEXT_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionNextClick();
			}
		});
		
		((JTextArea)currentForm.getComponentByName(NetworkConfigurationComponents.ID_HELP_TEXT_AREA)).setCaretPosition(0);
		this.setSize(430, 403);
		ProjectBuilder.addWindowCloseListener(this);
	}
	
	/**
	 * Populates all fields using the project data when this form is set to visible.
	 */
	public void setVisible(boolean isVisible) {
		if (isVisible == true) {
			populateFields();
		}
		super.setVisible(isVisible);
	}
	
	/**
	 * Reads information from the project and sets the fields in the form accordingly.
	 */
	private void populateFields() {
		// Get the project from the ProjectBuilder.
		project = ProjectBuilder.instance().getProject();
		if (project == null)
			project = Project.createProject();
		
		// Extract information from the project.
		String serverURI = project.getServerURI();
		if (serverURI == null)
			serverURI = "host.domain.edu";
		String[] serverURIComponents = serverURI.split(":");
		String serverAddress = serverURIComponents[0];
		Integer serverPort = 2080;
		if (serverURIComponents.length > 1)
				serverPort = Integer.parseInt(serverURIComponents[1]);
		
		int clientPort = project.getClientPort();
		if (clientPort == 0)
			clientPort = 2081;
		
		int sciappPort = project.getScienceAppPort();
		if (sciappPort == 0)
			sciappPort = 2082;
		
		// Populate the fields.
		currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).setText(serverAddress);
		serverPortSpinner.setValue(serverPort);
		clientPortSpinner.setValue(clientPort);
		sciappPortSpinner.setValue(sciappPort);
	}
	
	/**
	 * Updates the project properties using the values in the form fields. 
	 */
	private void updateProject() {
		String serverURI = currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).getText();
		serverURI += ":";
		serverURI += serverPortSpinner.getValue(); 
		project.setServerURI(serverURI);
		project.setClientPort((Integer)clientPortSpinner.getValue());
		project.setScienceAppPort((Integer)sciappPortSpinner.getValue());
	}
	
	/**
	 * Checks whether all input fields are correct and displays a dialog message if there are any errors.
	 * @return Returns true if all input fields are valid, false otherwise.
	 */
	private boolean validateInput() {
		boolean isValid = true;
		if (currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).getText().length() == 0) {
			isValid = false;
			ProjectBuilder.instance().showErrorDialog(this, "The server address must be specified.");
		}
		else if (currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).getText().equals("localhost") || 
				currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).getText().equals("127.0.0.1")) {
			String message = "It is not advised to use localhost as your project server address\n";
			message += "because some clients may not be able to connect.";
			ProjectBuilder.instance().showWarningDialog(this, message);
		}
		else if (currentForm.getTextField(NetworkConfigurationComponents.ID_SERVER_ADDRESS_FIELD).getText().equals("host.domain.edu")) {
			isValid = false;
			String message = "The project address \"host.domain.edu\" is given only as an example.\n";
			message += "Please enter your actual server address.";
			ProjectBuilder.instance().showErrorDialog(this, message);
		}
		else if (((Integer)serverPortSpinner.getValue()).intValue() == ((Integer)clientPortSpinner.getValue()).intValue() ||
				((Integer)serverPortSpinner.getValue()).intValue() == ((Integer)sciappPortSpinner.getValue()).intValue() ||
				((Integer)clientPortSpinner.getValue()).intValue() == ((Integer)sciappPortSpinner.getValue()).intValue()) {
			isValid = false;
			ProjectBuilder.instance().showErrorDialog(this, "No two port numbers can be the same.");
		}
		return isValid;
	}
	
	/**
	 * Controls the actions that occur when the previous button is clicked.
	 */
	private void selectActionPreviousClick() {
		updateProject();
		ProjectBuilder.instance().setProject(project);
		this.setVisible(false);
		ProjectBuilder.setRelativePosition(previousForm, this);
		previousForm.setVisible(true);
	}
	
	/**
	 * Controls the actions that occur when the next button is clicked.
	 */
	private void selectActionNextClick() {
		updateProject();
		if (validateInput() == true) {
			ProjectBuilder.instance().setProject(project);
			if (nextForm == null)
				nextForm = new ServerConfigurationForm(this);
			ProjectBuilder.setRelativePosition(nextForm, this);
			nextForm.setVisible(true);
			this.setVisible(false);
		}
	}
}
