package edu.wpi.prc.tools.builder.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;

import com.jeta.forms.components.panel.FormPanel;

import edu.wpi.prc.persistence.DatabaseType;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.tools.builder.ProjectBuilder;

/**
 * Allows the user to specify the type of transitioner and validator to use, as well as the minimum number of results for each work unit.
 * @author James Baldassari
 */
public class ServerConfigurationForm extends GUIForm {
	private static final long serialVersionUID = 1L;
	private GUIForm previousForm = null;
	private FormPanel currentForm = null;
	private GUIForm nextForm = null;
	private Project project = null;
	private SpinnerNumberModel remoteTransitionerPortSpinnerModel = new SpinnerNumberModel();
	private SpinnerNumberModel minNumResultsSpinnerModel = new SpinnerNumberModel();
	private JSpinner remoteTransitionerPortSpinner = null;
	private JSpinner minNumResultsSpinner = null;
	
	/**
	 * Constructor that takes a reference to the this form's parent form.
	 */
	public ServerConfigurationForm(GUIForm previousForm) {
		this.previousForm = previousForm;
		currentForm = new FormPanel("edu/wpi/prc/tools/builder/gui/forms/ServerConfigurationForm.jfrm");
		setLayout(new BorderLayout());
		add(currentForm, BorderLayout.CENTER);
		this.setTitle("PRC Project Builder: Step 5 of 7");
		
		remoteTransitionerPortSpinnerModel.setMinimum(1);
		remoteTransitionerPortSpinnerModel.setStepSize(1);
		remoteTransitionerPortSpinnerModel.setMaximum(65536);
		minNumResultsSpinnerModel.setMinimum(1);
		minNumResultsSpinnerModel.setStepSize(1);
		remoteTransitionerPortSpinner = currentForm.getSpinner(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_PORT_SPINNER);
		minNumResultsSpinner = currentForm.getSpinner(ServerConfigurationComponents.ID_MIN_NUM_RESULTS_SPINNER);
		
		remoteTransitionerPortSpinner.setModel(remoteTransitionerPortSpinnerModel);
		remoteTransitionerPortSpinner.setEditor(new JSpinner.NumberEditor(remoteTransitionerPortSpinner, "#####"));
		((JSpinner.DefaultEditor)remoteTransitionerPortSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		minNumResultsSpinner.setModel(minNumResultsSpinnerModel);
		((JSpinner.DefaultEditor)minNumResultsSpinner.getEditor()).getTextField().setHorizontalAlignment(JTextField.LEFT);
		

		// Add action listeners.
		
		currentForm.getRadioButton(ServerConfigurationComponents.ID_LOCAL_TRANSITIONER_RADIO).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectLocalTransitioner(event);
			}
		});
		
		currentForm.getRadioButton(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_RADIO).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectRemoteTransitioner(event);
			}
		});
		
		currentForm.getButton(ServerConfigurationComponents.ID_PREVIOUS_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionPreviousClick();
			}
		});
		
		currentForm.getButton(ServerConfigurationComponents.ID_NEXT_BUTTON).addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				selectActionNextClick();
			}
		});
		
		((JTextArea)currentForm.getComponentByName(ServerConfigurationComponents.ID_HELP_TEXT_AREA)).setCaretPosition(0);
		this.setSize(430, 422);
		ProjectBuilder.addWindowCloseListener(this);
	}
	
	/**
	 * Populates all fields using the project data when this form is set to visible.
	 */
	public void setVisible(boolean isVisible) {
		if (isVisible == true) {
			populateFields();
		}
		super.setVisible(isVisible);
	}
	
	/**
	 * Reads information from the project and sets the fields in the form accordingly.
	 */
	private void populateFields() {
		// Get the project from the ProjectBuilder.
		project = ProjectBuilder.instance().getProject();
		if (project == null)
			project = Project.createProject();
		
		// Extract information from the project.
		String transitionerURI = project.getTransitionerURI();
		String transitionerAddress = "";
		Integer transitionerPort = 0;
		if (transitionerURI == null)
			currentForm.getRadioButton(ServerConfigurationComponents.ID_LOCAL_TRANSITIONER_RADIO).setSelected(true);
		else {
			currentForm.getRadioButton(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_RADIO).setSelected(true);
			currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_LABEL).setEnabled(true);
			currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setEnabled(true);
			currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_PORT_LABEL).setEnabled(true);
			remoteTransitionerPortSpinner.setEnabled(true);
			String[] transitionerComponents = transitionerURI.split(":");
			transitionerAddress = transitionerComponents[0];
			if (transitionerComponents.length > 1)
				transitionerPort = Integer.parseInt(transitionerComponents[1]);
			else transitionerPort = 2083;
		}
		
		boolean useDefaultValidator = project.getDefaultValidator();
		if (useDefaultValidator == true)
			currentForm.getRadioButton(ServerConfigurationComponents.ID_DEFAULT_VALIDATOR_RADIO).setSelected(true);
		else
			currentForm.getRadioButton(ServerConfigurationComponents.ID_CUSTOM_VALIDATOR_RADIO).setSelected(true);
		
		Integer minNumResults = 1;
		if (project.getMinNumResults() > 1)
			minNumResults = project.getMinNumResults();
		
		// Populate the fields.
		currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setText(transitionerAddress);
		remoteTransitionerPortSpinner.setValue(transitionerPort);
		minNumResultsSpinner.setValue(minNumResults);
	}
	
	/**
	 * This method is called when the local transitioner type is changed.
	 * @param event The event that caused this action listener to be called.
	 */
	private void selectLocalTransitioner(ActionEvent event) {
		currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setText("");
		remoteTransitionerPortSpinner.setValue(0);
		currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_LABEL).setEnabled(false);
		currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setEnabled(false);
		currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_PORT_LABEL).setEnabled(false);
		remoteTransitionerPortSpinner.setEnabled(false);
	}
	
	/**
	 * This method is called when the transitioner type is changed.
	 * @param event The event that caused this action listener to be called.
	 */
	private void selectRemoteTransitioner(ActionEvent event) {
		String transitionerURI = project.getTransitionerURI();
		String transitionerAddress = "";
		Integer transitionerPort = 2083;
		if (transitionerURI != null) {
			String[] transitionerComponents = transitionerURI.split(":");
			transitionerAddress = transitionerComponents[0];
			if (transitionerComponents.length > 1)
				transitionerPort = Integer.parseInt(transitionerComponents[1]);
		}
		
		currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setText(transitionerAddress);
		remoteTransitionerPortSpinner.setValue(transitionerPort);
		
		currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_LABEL).setEnabled(true);
		currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).setEnabled(true);
		currentForm.getLabel(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_PORT_LABEL).setEnabled(true);
		remoteTransitionerPortSpinner.setEnabled(true);		
	}
	
	/**
	 * Updates the project properties using the values in the form fields. 
	 */
	private void updateProject() {
		String transitionerURI = currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).getText();
		transitionerURI += ":";
		transitionerURI += remoteTransitionerPortSpinner.getValue(); 
		
		project.setDefaultValidator(currentForm.getRadioButton(ServerConfigurationComponents.ID_DEFAULT_VALIDATOR_RADIO).isSelected());
		
		if (currentForm.getRadioButton(ServerConfigurationComponents.ID_LOCAL_TRANSITIONER_RADIO).isSelected() == true)
			project.setTransitionerURI(null);
		else
			project.setTransitionerURI(transitionerURI);
		
		project.setMinNumResults((Integer)minNumResultsSpinner.getValue());
	}
	
	/**
	 * Checks whether all input fields are correct and displays a dialog message if there are any errors.
	 * @return Returns true if all input fields are valid, false otherwise.
	 */
	private boolean validateInput() {
		boolean isValid = true;
		if (currentForm.getRadioButton(ServerConfigurationComponents.ID_LOCAL_TRANSITIONER_RADIO).isSelected() == false) {
			if (project.getDbType() == DatabaseType.HSQLDB) {
				isValid = false;
				String message = "A remote transitioner cannot be used in conjunction with an HSQLDB database.";
				ProjectBuilder.instance().showErrorDialog(this, message);
			}
			else if (currentForm.getTextField(ServerConfigurationComponents.ID_REMOTE_TRANSITIONER_ADDRESS_FIELD).getText().length() == 0) {
				isValid = false;
				String message = "The remote transitioner address must be specified if you want to use a remote transitioner.";
				ProjectBuilder.instance().showErrorDialog(this, message);
			}
			else if (((Integer)remoteTransitionerPortSpinner.getValue()).intValue() == Integer.parseInt(project.getServerURI().split(":")[1]) ||
					((Integer)remoteTransitionerPortSpinner.getValue()).intValue() == project.getClientPort() ||
					((Integer)remoteTransitionerPortSpinner.getValue()).intValue() == project.getScienceAppPort()) {
				isValid = false;
				String message = "The remote transitioner port cannot be the same as the project server port,\n";
				message += "the client port, or the science application port.";
				ProjectBuilder.instance().showErrorDialog(this, message);				
			}
		}
		else if ((currentForm.getRadioButton(ServerConfigurationComponents.ID_DEFAULT_VALIDATOR_RADIO).isSelected() == true) &&
				(project.getMinNumResults() > 1)) {
			String message = "Requiring more than one result per work unit when you are using the\n";
			message += "default validator will cause your project to be needlessly inefficient\n";
			message += "because the default validator ignores all results except the first.";
			ProjectBuilder.instance().showWarningDialog(this, message);
		}
		return isValid;
	}
	
	/**
	 * Controls the actions that occur when the previous button is clicked.
	 */
	private void selectActionPreviousClick() {
		updateProject();
		ProjectBuilder.instance().setProject(project);
		this.setVisible(false);
		ProjectBuilder.setRelativePosition(previousForm, this);
		previousForm.setVisible(true);
	}
	
	/**
	 * Controls the actions that occur when the next button is clicked.
	 */
	private void selectActionNextClick() {
		updateProject();
		if (validateInput() == true) {
			ProjectBuilder.instance().setProject(project);
			if (nextForm == null)
				nextForm = new SpotCheckConfigurationForm(this);
			ProjectBuilder.setRelativePosition(nextForm, this);
			nextForm.setVisible(true);
			this.setVisible(false);
		}
	}
}
