package edu.wpi.prc.tools.shutdown;

import java.io.IOException;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import edu.wpi.prc.project.MissingConfigurationFileException;
import edu.wpi.prc.project.Project;
import edu.wpi.prc.util.password.PasswordReader;

/**
 * Reads the proejct configuration and shuts down the project 
 * server or client by sending an XML-RPC request.
 * @author James Baldassari
 */
public class ShutdownController {
	private static final int EXIT_SUCCESS = 0;
	private static final int EXIT_ERROR = 1;
	private static Project project = null;
	private static XmlRpcClient xmlrpcClient = null;
	
	/**
	 * Executes the "server.shutdown" RPC on the project server using the given server address and password.
	 * @param serverAddress The server address.
	 * @param password The project password.
	 * @return Returns true if the XML-RPC call was successful, the password was correct, 
	 * and the server began shutting down; false otherwise.
	 */
	private static boolean shutdownServer(String serverAddress, String password) {
		boolean success = false;
		String serverURI = "http://" + serverAddress + "/RPC2";
		try {
			// Instantiate an XML-RPC client.
			xmlrpcClient = new XmlRpcClient(serverURI);
			// Build Vector of parameters.
			Vector<String> params = new Vector<String>();
			params.add(password);
			success = sendRPC("server.shutdown", params);
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + serverURI + "\" from the project configuration file is invalid.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Executes the "server.kill" RPC on the project server using the given server address and password.
	 * This RPC does not wait for all transactions to be completed before shutting down the server, 
	 * and should only be used in the event that one or more RPCs are "stuck," preventing the server 
	 * from shutting down normally.
	 * @param serverAddress The server address.
	 * @param password The project password.
	 * @return Returns true if the XML-RPC call was successful, the password was correct, 
	 * and the server began shutting down; false otherwise.
	 */
	private static boolean emergencyShutdown(String serverAddress, String password) {
		boolean success = false;
		String serverURI = "http://" + serverAddress + "/RPC2";
		try {
			// Instantiate an XML-RPC client.
			xmlrpcClient = new XmlRpcClient(serverURI);
			// Build Vector of parameters.
			Vector<String> params = new Vector<String>();
			params.add(password);
			success = sendRPC("server.kill", params);
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + serverURI + "\" from the project configuration file is invalid.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Executes the "transitioner.shutdown" RPC on the project server using the given server address and password.
	 * @param transitionerAddress The remote transitioner address.
	 * @param password The project password.
	 * @return Returns true if the XML-RPC call was successful, the password was correct, 
	 * and the server began shutting down; false otherwise.
	 */
	private static boolean shutdownTransitioner(String transitionerAddress, String password) {
		boolean success = false;
		String transitionerURI = "http://" + transitionerAddress + "/RPC2";
		try {
			// Instantiate an XML-RPC client.
			xmlrpcClient = new XmlRpcClient(transitionerURI);
			// Build Vector of parameters.
			Vector<String> params = new Vector<String>();
			params.add(password);
			success = sendRPC("transitioner.shutdown", params);
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + transitionerURI + "\" from the project configuration file is invalid.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Executes the "client.shutdown" RPC on the project client.
	 * @param port The port on which the client is listening for XML-RPC connections.
	 * @return Returns true if the RPC was executed successfully, false otherwise.
	 */
	private static boolean shutdownClient(int port) {
		boolean success = false;
		String clientURI = "http://localhost:" + port + "/RPC2";
		try {
			// Instantiate an XML-RPC client.
			xmlrpcClient = new XmlRpcClient(clientURI);
			success = sendRPC("client.shutdown", new Vector());
		} catch (MalformedURLException e) {
			System.out.println("The URL \"" + clientURI + "\" is invalid.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Executes the given XML-RPC command with the given parameters.
	 * @param cmd The command to send.
	 * @param params The parameters to pass.
	 * @return True if the command was sent successfully, false otherwise.
	 */
	private static boolean sendRPC(String cmd, Vector params) {
		boolean success = false;
		System.out.println("Sending " + cmd.split("\\.")[0] + " shutdown command.");
		try {
			Object result = xmlrpcClient.execute(cmd, params);
			try {
				success = (Boolean)result;
			} catch (Exception e) {
				success = false;
			}
		} catch (XmlRpcException e) {
			System.out.println("Error sending " + cmd.split("\\.")[0] + " shutdown command: " + e);
			System.exit(EXIT_ERROR);
		} catch (IOException e) {
			System.out.println("Could not connect to the " + cmd.split("\\.")[0] + ".  It may already be shut down.");
			System.exit(EXIT_ERROR);
		}
		return success;
	}
	
	/**
	 * Determines which process to shut down by parsing
	 * command-line arguments.  If the user wants to 
	 * shut down the server, the user is prompted for 
	 * the project password.  The project configuration 
	 * file is parsed to determine address/port, then
	 * the XML-RPC request is sent.
	 * @param args Command-line arguments.
	 */
	public static void main(String[] args) {
		// For some reason the Apache XML-RPC library sends output to System.err.
		// It's really annoying, so just redirect System.err into oblivion.
		System.setErr(new PrintStream(new PipedOutputStream()));
		
		if (args.length != 1) {
			System.out.println("You must specify either \"server\" or \"client\" in the command-line arguments.");
			System.exit(EXIT_ERROR);
		}
		
		// If the "server" argument was passed, try to shut down the server.
		if (args[0].equals("server")) {
			try {
				project = Project.readFromFile();
				String serverURI = project.getServerURI();				
				if (serverURI == null) {
					System.out.println("Error: Server address not found in project configuration file.");
					System.exit(EXIT_ERROR);
				}
				else {
					PasswordReader pr = new PasswordReader();
					try {
						// Read in the project password.
						String password = pr.readLine();
						// Try to shut down the server.
						while (shutdownServer(serverURI, password) == false) {
							System.out.println("The password you entered was incorrect.  Please enter the password again.");
							password = pr.readLine();
						}
						System.out.println("The project server is shutting down.");
						System.out.println("Please check the server log file to determine when the server has finished shutting down.");
						System.exit(EXIT_SUCCESS);
					} catch (IOException e) {
						System.out.println("Error reading user input.");
						System.exit(EXIT_ERROR);
					}
				}
			} catch (MissingConfigurationFileException e) {
				String message = "Project configuration file not found.  Please verify that ";
				message += "the project_server.properties file is in the \"cfg\" directory.";
				System.out.println(message);
				System.exit(EXIT_ERROR);
			}
		}
		
		// If the "emergency" argument was passed, try to shut down the server without waiting for all transactions to complete.
		else if (args[0].equals("emergency")) {
			try {
				project = Project.readFromFile();
				String serverURI = project.getServerURI();				
				if (serverURI == null) {
					System.out.println("Error: Server address not found in project configuration file.");
					System.exit(EXIT_ERROR);
				}
				else {
					PasswordReader pr = new PasswordReader();
					try {
						// Read in the project password.
						String password = pr.readLine();
						// Try to shut down the server.
						while (emergencyShutdown(serverURI, password) == false) {
							System.out.println("The password you entered was incorrect.  Please enter the password again.");
							password = pr.readLine();
						}
						System.out.println("The project server is shutting down.");
						System.out.println("Please check the server log file to determine when the server has finished shutting down.");
						System.exit(EXIT_SUCCESS);
					} catch (IOException e) {
						System.out.println("Error reading user input.");
						System.exit(EXIT_ERROR);
					}
				}
			} catch (MissingConfigurationFileException e) {
				String message = "Project configuration file not found.  Please verify that ";
				message += "the project_server.properties file is in the \"cfg\" directory.";
				System.out.println(message);
				System.exit(EXIT_ERROR);
			}
		}
		
		else if (args[0].equals("transitioner")) {
			try {
				project = Project.readFromFile();
				String transitionerURI = project.getTransitionerURI();				
				if (transitionerURI == null) {
					System.out.println("Error: Remote transitioner address not found in project configuration file.");
					System.exit(EXIT_ERROR);
				}
				else {
					PasswordReader pr = new PasswordReader();
					try {
						// Read in the project password.
						String password = pr.readLine();
						// Try to shut down the server.
						while (shutdownTransitioner(transitionerURI, password) == false) {
							System.out.println("The password you entered was incorrect.  Please enter the password again.");
							password = pr.readLine();
						}
						System.out.println("The remote transitioner is shutting down.");
						System.exit(EXIT_SUCCESS);
					} catch (IOException e) {
						System.out.println("Error reading user input.");
						System.exit(EXIT_ERROR);
					}
				}
			} catch (MissingConfigurationFileException e) {
				String message = "Project configuration file not found.  Please verify that ";
				message += "the project_server.properties file is in the \"cfg\" directory.";
				System.out.println(message);
				System.exit(EXIT_ERROR);
			}
		}
		
		// If the "client" argument was passed, try to shut down the client.
		else if (args[0].equals("client")) {
			try {
				project = Project.readFromFile();
				int clientPort = project.getClientPort();
				if (clientPort == 0) {
					System.out.println("Error: Client port not found in project configuration file.");
					System.exit(	EXIT_ERROR);
				}
				else {
					if (shutdownClient(clientPort) == true) {
						System.out.println("The project client is shutting down.");
						System.exit(EXIT_SUCCESS);
					}
				}
			} catch (MissingConfigurationFileException e) {
				String message = "Project configuration file not found.  Please verify that ";
				message += "the project_client.properties file is in the \"cfg\" directory.";
				System.out.println(message);
				System.exit(EXIT_ERROR);
			}
		}
		
		else {
			System.out.println("Invalid argument: \"" + args[0] + "\"");
			System.out.println("You must specify either \"server\" or \"client\" in the command-line arguments.");
			System.exit(EXIT_ERROR);
		}
	}
}
